# import uuid
# from celery import shared_task

# from app.settings import (
#     NOTICE_SERVICE_BASE_URL,
#     RECOVERY_SERVICE_BASE_URL,
#     REDIS,
#     S3_BUCKET_ENDPOINT,
#     S3_BUCKET_NAME,
# )
# from app.tasks.notice.helpers import (
#     generate_notice_for_all_addresses,
#     logs_tracker,
#     map_formatter,
# )
# import logging
# import redis
# import json

# from app.tasks.notice.loan_service import get_loan_data
# from app.tasks.notice.notice_service import (
#     get_notice_generation_data,
#     insert_into_db,
#     upload_notices_to_s3,
# )

# logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# failed_notice_id_list = []


# @shared_task(bind=True, name="generate_physical_notices", ignore_result=False, acks_late=False)
# def generate_physical_notices(self, *args, **kwargs):
#     logger.info("notice.generate_physical_notices")
#     logs_loan_id = kwargs["linked_loan"]["linked_loan_id"] if kwargs.get("linked_loan") else kwargs.get("loan_ids")[0]
#     logger.info(f"notice.generate_physical_notices.loan_id :: {logs_loan_id}")
#     logger.info(f"notice.generate_physical_notices.batch_id :: {kwargs['batch_id']}")

#     try:
#         redis_instance = redis.Redis(host=REDIS["HOST"], port=REDIS["PORT"])
#         notice_kwargs = json.loads(redis_instance.get(f"notice_kwargs_{kwargs['batch_id']}").decode("utf-8"))
#         redis_instance.close()
#         # In case key is not present in redis
#         if not notice_kwargs:
#             logger.error("notice.generate_physical_notices.exception: notice_kwargs not found in redis")
#             raise Exception("notice.generate_physical_notices.exception: notice_kwargs not found in redis")

#         kwargs = {**kwargs, **notice_kwargs}
#         logger.debug(f"notice.generate_physical_notices.kwargs : {kwargs}")

#         local_batch_directory_path = kwargs["local_batch_directory_path"]
#         local_draft_file_path = kwargs["local_draft_file_path"]
#         local_doc_directory_path = kwargs["local_doc_directory_path"]
#         local_pdf_directory_path = kwargs["local_pdf_directory_path"]
#         local_qr_code_directory_path = kwargs["local_qr_code_directory_path"]
#         local_merged_pdf_directory_path = kwargs["local_merged_pdf_directory_path"]
#         local_merged_pdf_zip_file_path = kwargs["local_merged_pdf_zip_file_path"]
#         preview = kwargs.get("preview", False)
#         request_id = str(uuid.uuid4())

#         # For linked loans
#         if kwargs.get("linked_loan"):
#             linked_loan_dict = kwargs["linked_loan"]
#             linked_loan_id = linked_loan_dict.get("linked_loan_id", "")
#             loan_ids = linked_loan_dict.get("loan_ids", [])

#             if linked_loan_id in (None, "") or len(loan_ids) == 0:
#                 raise Exception(f"notice.generate_physical_notices.exception: Empty loan_ids for this linked loan")
#         # For normal loans
#         else:
#             loan_ids = kwargs["loan_ids"]

#         notice_generation_kwargs = {
#             "company_id": kwargs["company_id"],
#             "draft_id": kwargs["draft_id"],
#             "draft_details": kwargs["draft_details"],
#             "loan_ids": loan_ids,
#             "linked_loan_id": linked_loan_id if kwargs.get("linked_loan") else "",
#             "allocation_month": kwargs["allocation_month"],
#             "author": kwargs["author"],
#             "role": kwargs["role"],
#             "batch_number": kwargs["batch_number"],
#             "batch_id": kwargs["batch_id"],
#             "document_type": kwargs["document_type"],
#             "mode": kwargs["mode"],
#             "type_of_task": "physical",
#             "preview": preview,
#             "auth_token": kwargs["token"],
#             "BUCKET_NAME": S3_BUCKET_NAME,
#             "S3_BUCKET_ENDPOINT": S3_BUCKET_ENDPOINT,
#             "RECOVERY_SERVICE_BASE_URL": RECOVERY_SERVICE_BASE_URL,
#             "local_pdf_directory_path": local_pdf_directory_path,
#             "local_doc_directory_path": local_doc_directory_path,
#             "local_batch_directory_path": local_batch_directory_path,
#             "local_draft_file_path": local_draft_file_path,
#             "local_qr_code_directory_path": local_qr_code_directory_path,
#             "local_merged_pdf_directory_path": local_merged_pdf_directory_path,
#             "local_merged_pdf_zip_file_path": local_merged_pdf_zip_file_path,
#             "NOTICE_SERVICE_BASE_URL": NOTICE_SERVICE_BASE_URL,
#             "use_tracking_ids": kwargs.get("use_tracking_ids", False),
#             "physical_notice_trackings_ids_redis_key": kwargs.get("physical_notice_trackings_ids_redis_key", ""),
#             "user": kwargs["user"],
#             "is_linked_loan": kwargs["is_linked_loan"],
#             "request_id": request_id,
#             "draft_fields_redis_key": kwargs["draft_fields_redis_key"],
#             "company": kwargs["company"],
#             "initial_notice_id_redis_key": kwargs["initial_notice_id_redis_key"],
#             "enable_qrcode": kwargs["enable_qrcode"],
#             "merged_lot_size": kwargs.get("merged_lot_size"),
#         }

#         # To make filename based on s3_link_uuid which is actually batch_id
#         if preview:
#             notice_generation_kwargs["s3_link_uuid"] = kwargs["s3_link_uuid"]

#         generate_physical_notice(**notice_generation_kwargs)

#         message = (
#             "Linked loan physical notice generated" if kwargs.get("linked_loan_id") else "Physical notice generated"
#         )

#         logs_tracker_req_body = {
#             "response": {"message": f"{message}"},
#             "kwargs": {
#                 "batch_id": kwargs["batch_id"],
#                 "author": kwargs["user"]["email"],
#                 "company_id": kwargs["company"]["company_id"],
#                 "loan_id": logs_loan_id,
#                 "user": kwargs["user"],
#                 "company": kwargs["company"],
#             },
#             "status_code": 200,
#             "status": "SUCCESS",
#         }
#     except Exception as e:
#         logger.error(f"generate_physical_notices.exception :: {str(e)}")
#         logs_tracker_req_body = {
#             "response": {"message": str(e), "notice_id_list": failed_notice_id_list},
#             "kwargs": {
#                 "batch_id": kwargs["batch_id"],
#                 "author": kwargs["user"]["email"],
#                 "company_id": kwargs["company"]["company_id"],
#                 "loan_id": logs_loan_id,
#                 "user": kwargs["user"],
#                 "company": kwargs["company"],
#             },
#             "status_code": 400,
#             "status": "FAIL",
#         }
#     finally:
#         logs_tracker(logs_tracker_req_body, kwargs.get("X-Request-Id"))
#         failed_notice_id_list.clear()


# def generate_physical_notice(**kwargs):
#     logger.info("notice.generate_physical_notice")
#     logger.info(f"notice.generate_physical_notice.batch_id :: {kwargs['batch_id']}")

#     logger.debug(f"notice.generate_physical_notice.kwargs: {kwargs}")
#     preview = kwargs.get("preview", False)
#     kwargs["X-Request-Id"] = kwargs["request_id"]
#     failed_tracking_id_list = []

#     try:
#         redis_instance = redis.Redis(host=REDIS["HOST"], port=REDIS["PORT"])
#         draft_fields = redis_instance.get(kwargs["draft_fields_redis_key"]).decode("utf-8")

#         if not draft_fields:
#             raise Exception("tasks.generate_physical_notice.exception: Unable to fetch draft details from redis")

#         draft_fields = json.loads(draft_fields)

#         draft_details = kwargs.get("draft_details", {})
#         kwargs["draft_name"] = draft_details.get("draft_name")
#         kwargs["notice_type"] = draft_details.get("notice_type")
#         kwargs["case_type"] = draft_details.get("case_type")
#         kwargs["is_vernacular"] = draft_details.get("is_vernacular")
#         kwargs["language"] = draft_details.get("language")
#         kwargs["draft_fields"] = draft_fields.get(kwargs.get("draft_id"))
#         kwargs["is_in_case"] = False

#         if kwargs["case_type"] not in ("", None):
#             kwargs["is_in_case"] = True

#         loan_data = get_loan_data(**kwargs)

#         if not loan_data:
#             logger.error("generate_physical_notice.get_loan_data_res.err : Can't find loan data")
#             raise Exception(
#                 "generate_physical_notice.get_loan_data_res.err: Empty loan_data response for the given loan"
#             )
#         if isinstance(loan_data, str):
#             logger.error(f"generate_physical_notice.get_loan_data_res.err : {loan_data}")
#             raise Exception(f"generate_physical_notice.get_loan_data_res.err: {loan_data}")
#         if len(kwargs.get("loan_ids", [])) != len(loan_data["company_data"]):
#             logger.error(f"generate_physical_notice.get_loan_data_res.err : Can't find data for all the loans")
#             raise Exception(f"generate_physical_notice.get_loan_data_res.err: Cant find details for all the loans")

#         kwargs["loan_data"] = loan_data["company_data"]
#         kwargs["loan_address_map"] = {}

#         # To get the key value pairs for merging with the notice document, where key is the merge field with corresponding value from the db records
#         notice_data = get_notice_generation_data(**kwargs)

#         if isinstance(notice_data, str):
#             logger.debug(f"generate_physical_notice.get_notice_generation_data.err : {notice_data}")
#             raise Exception(f"generate_physical_notice.err : {notice_data}")
#         else:
#             kwargs["notice_data"] = notice_data
#             kwargs["filenames"] = []
#             kwargs["tracking_ids"] = []
#             kwargs["local_pdf_file_names"] = []
#             kwargs["primary_addresses"] = []
#             kwargs["address_file_map"] = {}
#             kwargs["loan_allocation_map"] = get_loan_allocation_month_mapping(kwargs["loan_data"])

#             response, notice_id_list, failed_tracking_id_list = generate_notice_for_all_addresses(**kwargs)

#             global failed_notice_id_list
#             failed_notice_id_list = notice_id_list
#             kwargs["notice_id_list"] = notice_id_list

#             if response != "success":
#                 raise Exception(response)

#             if not preview:
#                 changed_dt = notice_data.get("changed_dt", {})
#                 db_data = {
#                     "applicant_type": changed_dt["applicant_type"],
#                     "address_index": changed_dt["address_index"],
#                     "batch_number": kwargs.get("batch_number", ""),
#                     "notice_reference_number": changed_dt["notice_reference_number"],
#                     "applicant_notice_reference_number": changed_dt["applicant_notice_reference_number"],
#                     "co_applicant_notice_reference_number": changed_dt["co_applicant_notice_reference_number"],
#                     "notice_variables_data": changed_dt,
#                 }
#                 kwargs["db_data"] = db_data
#                 kwargs["files_s3_map"] = {}

#             upload_response = upload_notices_to_s3(**kwargs)

#             if not preview:
#                 logger.debug(f"generate_physical_notice.upload_response: {upload_response}")
#                 kwargs["loan_address_file_s3_map"] = {}
#                 loan_address_file_s3_map_list = map_formatter(
#                     kwargs["loan_address_map"],
#                     kwargs["address_file_map"],
#                     kwargs["files_s3_map"],
#                 )
#                 kwargs["loan_address_file_s3_map_list"] = loan_address_file_s3_map_list
#                 kwargs["s3_link_ids"] = upload_response["s3_link_ids"]
#                 kwargs["s3_object_urls"] = upload_response["s3_object_urls"]

#                 for data in loan_address_file_s3_map_list:
#                     if str(data["notice_id"]) == str(kwargs["notice_id_list"][0]):
#                         kwargs["primary_notice_data"] = {
#                             "s3_link": data["s3_link"],
#                             "s3_link_uuid": data["s3_link_uuid"],
#                         }
#                 insert_into_db(**kwargs)

#     except Exception as e:
#         if failed_tracking_id_list:
#             failed_tracking_ids_key = f'failed_tracking_id_{kwargs.get("batch_id","")}'
#             redis_instance = redis.Redis(host=REDIS["HOST"], port=REDIS["PORT"])
#             redis_instance.lpush(failed_tracking_ids_key, *failed_tracking_id_list)
#             redis_instance.close()
#         raise Exception(f"generate_physical_notice.exception :: {str(e)}")


# def get_loan_allocation_month_mapping(loan_data):
#     loan_allocation_map = {}
#     logger.info("generate_physical_notice.get_loan_allocation_month_mapping")
#     for loan in loan_data:
#         loan_id = loan["loan_id"]
#         allocation_month = loan["question_dict"]["allocation_month"]
#         loan_allocation_map[loan_id] = allocation_month
#     return loan_allocation_map



"""
cg_celery.py
Usage: Celery
"""
import logging

from celery import Celery

from logging import Filter
from cg_logger import CustomFormatter
from app.settings import DTMF_RATE_LIMIT, CELERY_VISIBILITY_SETTINGS_ENABLED, CELERY_VISIBILITY_TIMEOUT

LOGGER_CONFIGURATION = (
    '{"timestamp": "%(asctime)s", "level": "%(levelname)s", "module": "%(module)s", '
    '"path": "%(pathname)s", "function": "%(funcName)s", "line":"%(lineno)d", '
    '"request_id": "%(request_id)s", "message": "%(message)s", "json_message": %(json_message)s}'
)


class QueueLogFilter(Filter):
    def filter(self, record):
        # from celery import current_task as celery_task
        # if celery_task == None:
        #     return False
        # print(f"current_task :: {celery_task}")
        # print(f"current_task.dir :: {dir(celery_task)}")
        # print(f"current_task.request :: {celery_task.request}")
        # print(f"current_task.Request :: {celery_task.Request}")

        if record.module in (
            "connection",
            "tasks",
            "report",
            "execution",
            "delegator",
            "callback",
            "service",
        ):
            return True
        return False


def make_celery(app):
    """
    Make celery instance to register tasks and executes them.
    """
    celery = Celery(
        app.import_name,
        backend=app.config["CELERY_RESULT_BACKEND"],
        broker=app.config["CELERY_BROKER_URL"],
    )
    celery.conf.update(app.config)

    if CELERY_VISIBILITY_SETTINGS_ENABLED:
        celery.conf.BROKER_TRANSPORT_OPTIONS = {"visibility_timeout": CELERY_VISIBILITY_TIMEOUT}
        print(f"\n CUSTOM CELERY SETTINGS || celery_visibility_timeout : {CELERY_VISIBILITY_TIMEOUT} \n")
    logger = logging.getLogger()
    sh = logging.StreamHandler()

    formatter = CustomFormatter(LOGGER_CONFIGURATION, None, "%", False)
    sh.setFormatter(formatter)
    sh.addFilter(QueueLogFilter())

    logger.setLevel(logging.DEBUG)
    logger.addHandler(sh)

    celery.autodiscover_tasks(
        [
            "app.tasks.result_tasks",
            "app.tasks.communication_result",
            "app.tasks.tasks",
            "app.tasks.delegator",
            "app.tasks.execution",
            "app.tasks.report",
        ],
        force=True,
    )
    celery.control.rate_limit("indiapost_tracking", "40/s")
    celery.control.rate_limit("communication", "100/s")
    celery.control.rate_limit("communication_dtmf_ivr", DTMF_RATE_LIMIT)
    celery.control.rate_limit("add_ecourt_case", "2/s")
    celery.control.rate_limit("update_ecourt_case", "2/s")
    celery.control.rate_limit("fetch_ecourt_case_orders", "2/s")
    celery.control.rate_limit("lat_long_conversion", "2/s")
    celery.worker_prefetch_multiplier = 1
    return celery



from datetime import datetime
from os import getenv
from ast import literal_eval

from dotenv import load_dotenv, find_dotenv

load_dotenv(find_dotenv())

# App Variables
APP_NAME = "queue_service"
ENV = getenv("ENV", "PRODUCTION").lower()
BASE_ROUTE = getenv("BASE_ROUTE")
QUEUE_NAME = getenv("QUEUE_NAME")
LOG_LEVEL = getenv("LOG_LEVEL", "DEBUG")
MAX_CONTENT_LENGTH = None
# URLS
USER_SERVICE_BASE_URL = getenv("USER_SERVICE_BASE_URL")
COMMUNICATION_SERVICE_BASE_URL = getenv("COMMUNICATION_SERVICE_BASE_URL")
SCRAPE_SERVICE_BASE_URL = getenv("SCRAPE_SERVICE_BASE_URL")
NOTICE_SERVICE_BASE_URL = getenv("NOTICE_SERVICE_BASE_URL")
QUEUE_SERVICE_BASE_URL = getenv("QUEUE_SERVICE_BASE_URL")
RECOVERY_SERVICE_BASE_URL = getenv("RECOVERY_SERVICE_BASE_URL")
LITIGATION_SERVICE_BASE_URL = getenv("LITIGATION_SERVICE_BASE_URL")
HEROFINCORP_URL = getenv("HEROFINCORP_URL")
UI_SERVICE_BASE_URL = getenv("UI_SERVICE_BASE_URL")
MANHATTAN_SERVICE_BASE_URL = getenv("MANHATTAN_SERVICE_BASE_URL")
ATTEMPTED_DATE = datetime.strptime(getenv("ATTEMPTED_DATE"), "%Y-%m-%d").date()


# DB
POSTGRES = {
    "HOST": getenv("DB_HOST"),
    "PORT": int(getenv("DB_PORT")),
    "USER": getenv("QUEUE_DB_USER"),
    "PASSWORD": getenv("QUEUE_DB_PASSWORD"),
    "NAME": getenv("QUEUE_DB_NAME"),
    "ENABLE_DB_READ_REPLICA": getenv("ENABLE_DB_READ_REPLICA"),
    "READ_REPLICA_DB_HOST": getenv("READ_REPLICA_DB_HOST"),
    "READ_REPLICA_DB_PORT": getenv("DB_PORT"),
}

# CAMPAIGN DB
CAMPAIGN_POSTGRES = {
    "HOST": getenv("DB_HOST"),
    "PORT": int(getenv("DB_PORT")),
    "USER": getenv("CAMPAIGN_DB_USER"),
    "PASSWORD": getenv("CAMPAIGN_DB_PASSWORD"),
    "NAME": getenv("CAMPAIGN_DB_NAME"),
    "ENABLE_DB_READ_REPLICA": getenv("ENABLE_DB_READ_REPLICA"),
    "READ_REPLICA_DB_HOST": getenv("READ_REPLICA_DB_HOST"),
    "READ_REPLICA_DB_PORT": getenv("DB_PORT"),
}
# RECOVERY DB
RECOVERY_POSTGRES = {
    "HOST": getenv("RECOVERY_DB_HOST"),
    "PORT": int(getenv("DB_PORT")),
    "USER": getenv("RECOVERY_DB_USER"),
    "PASSWORD": getenv("RECOVERY_DB_PASSWORD"),
    "NAME": getenv("RECOVERY_DB_NAME"),
    "ENABLE_DB_READ_REPLICA": getenv("ENABLE_RECOVERY_DB_READ_REPLICA"),
    "READ_REPLICA_DB_HOST": getenv("RECOVERY_READ_REPLICA_DB_HOST"),
    "READ_REPLICA_DB_PORT": getenv("DB_PORT"),
}

COMM_POSTGRES = {
    "HOST": getenv("DB_HOST"),
    "PORT": int(getenv("DB_PORT")),
    "USER": getenv("COMMUNICATION_DB_USER"),
    "PASSWORD": getenv("COMMUNICATION_DB_PASSWORD"),
    "NAME": getenv("COMMUNICATION_DB_NAME"),
    "ENABLE_DB_READ_REPLICA": getenv("ENABLE_DB_READ_REPLICA"),
    "READ_REPLICA_DB_HOST": getenv("READ_REPLICA_DB_HOST"),
    "READ_REPLICA_DB_PORT": getenv("DB_PORT"),
}

# Redis
REDIS = {"HOST": getenv("REDIS_HOST"), "PORT": int(getenv("REDIS_PORT"))}


# Celery Variables
CELERY_BROKER_URL = getenv("CELERY_REDIS_URL")
CELERY_RESULT_BACKEND = getenv("CELERY_REDIS_URL")
CELERY_TASK_SERIALIZER = "json"
CELERY_CREATE_MISSING_QUEUES = True
CELERYD_HIJACK_ROOT_LOGGER = True
CELERY_ALWAYS_EAGER = False
CELERY_TASK_RESULT_EXPIRES = 18000
CELERYD_TIME_LIMIT = 86400
CELERY_ACKS_LATE = True
DTMF_RATE_LIMIT = getenv("DTMF_RATE_LIMIT")
DTMF_CHUNK_SIZE = int(getenv("DTMF_CHUNK_SIZE", 100))

# Campaign
CAMPAIGN_MAX_LIMIT = getenv("CAMPAIGN_MAX_LIMIT")
# AWS
# s3
AWS_ACCESS_KEY_ID = getenv("AWS_S3_ACCESS_KEY")
AWS_SECRET_ACCESS_KEY = getenv("AWS_S3_SECRET_KEY")

# Volume Directory
VOLUME_MOUNT_DIRECTORY = getenv("VOLUME_MOUNT_DIRECTORY")
NOTICE_VOLUME_MOUNT_DIRECTORY = getenv("NOTICE_VOLUME_MOUNT_DIRECTORY")

# sendgrid template ids
SENDGRID_API_KEY = getenv("SENDGRID_API_KEY")
SENDGRID_TEMPLATE_ID = getenv("SENDGRID_TEMPLATE_ID")
BATCH_SENDGRID_TEMPLATE_ID = getenv("BATCH_SENDGRID_TEMPLATE_ID")
CAMPAIGN_SENDGRID_TEMPLATE_ID = getenv("CAMPAIGN_SENDGRID_TEMPLATE_ID")
REPORT_SENDGRID_TEMPLATE_ID = getenv("REPORT_SENDGRID_TEMPLATE_ID")
COMPLETION_REPORT_SENDGRID_ID = getenv("COMPLETION_REPORT_SENDGRID_ID")
AI_COMPLETION_REPORT_SENDGRID_ID = getenv("AI_COMPLETION_REPORT_SENDGRID_ID")

# HEROFIN CORP
HEROUSER = getenv("HEROUSER")
HEROPASS = getenv("HEROPASS")


S3_LITIGATION_REPORTS_DIRECTORY = getenv("S3_LITIGATION_REPORTS_DIRECTORY")
LITIGATION_REPORT_BATCH_SIZE = int(getenv("LITIGATION_REPORT_BATCH_SIZE", 1000))
S3_LITIGATION_ORDERS_DIRECTORY = getenv("S3_LITIGATION_ORDERS_DIRECTORY")
S3_BUCKET_NAME = getenv("CG_BUCKET_NAME")
EXPORTS_BUCKET_NAME = getenv("EXPORTS_BUCKET_NAME")
S3_BUCKET_ENDPOINT = getenv("CG_CLOUDFRONT_ENDPOINT")
S3_CLOUDFRONT_BUCKET_ENDPOINT = getenv("CG_CLOUDFRONT_ENDPOINT_1")
DOMAIN_NAME = getenv("DOMAIN_NAME")

SERVICE_NAME = getenv("SERVICE_NAME")
MODULE_NAME = getenv("MODULE_NAME")
CUSTOM_HEADER_RPC_SECRET_KEY = getenv("CUSTOM_HEADER_RPC_SECRET_KEY")
# SMS AUTH KEY
INFOBIP_SMS_AUTH = getenv("INFOBIP_SMS_AUTH")

ELASTIC_APM = {
    # Set the required service name. Allowed characters:
    # a-z, A-Z, 0-9, -, _, and space
    "SERVICE_NAME": getenv("SERVICE_NAME"),
    # Use if APM Server requires a secret token
    "SECRET_TOKEN": getenv("APM_SECRET_TOKEN"),
    # Set the custom APM Server URL (default: http://localhost:8200)
    "SERVER_URL": getenv("APM_SERVER_URL"),
    # Set the service environment
    "ENVIRONMENT": getenv("ENV"),
    "LOG_LEVEL": getenv("LOG_LEVEL", "info").lower(),
    "VERIFY_SERVER_CERT": bool(getenv("APM_VERIFY_SERVER_CERT", "False")),
    "TRANSACTION_IGNORE_URLS": ["*/healthz"],
    "LOCAL_VAR_DICT_MAX_LENGTH": int(getenv("APM_LOCAL_VAR_DICT_MAX_LENGTH", "25")),
    "LOCAL_VAR_MAX_LENGTH": int(getenv("APM_LOCAL_VAR_MAX_LENGTH", "1000")),
    "CAPTURE_BODY": getenv("APM_CAPTURE_BODY", "off"),
    "SPAN_STACK_TRACE_MIN_DURATION": getenv("APM_CAPTURE_BODY", 0),
}

MERGE_NOTICE_CLIENT_ID = getenv("MERGE_NOTICE_CLIENT_ID")
MERGE_NOTICE_TOPIC = getenv("MERGE_NOTICE_TOPIC")
NUMBER_OF_PARTITIONS_FOR_MERGE_NOTICE_TOPIC = int(getenv("NUMBER_OF_PARTITIONS_FOR_MERGE_NOTICE_TOPIC"))

BCC_EMAILS = literal_eval(getenv("BCC_EMAILS", "[]"))

# DSC Variables
TRUECOPY_BASE_URL = getenv("TRUECOPY_BASE_URL")
TRUECOPY_API_KEY = getenv("TRUECOPY_API_KEY")
TRUECOPY_ID = getenv("TRUECOPY_ID")
TRUECOPY_PWD = getenv("TRUECOPY_PWD")
TRUECOPY_ACCESS_ID = getenv("TRUECOPY_ACCESS_ID")
API_TOKEN = getenv("API_TOKEN")
BASE_URL = getenv("BASE_URL")
COUNTRY_WISE_MAPPING = literal_eval(getenv("COUNTRY_WISE_MAPPING", "{}"))
EXPORT_BATCH_SIZE = int(getenv("EXPORT_BATCH_SIZE", "500"))
EXPORT_SERVICE_BASE_URL = getenv("EXPORT_SERVICE_BASE_URL")
AWS_DEFAULT_REGION = getenv("AWS_DEFAULT_REGION")
AWS_S3_ACCESS_KEY = getenv("AWS_S3_ACCESS_KEY")
AWS_S3_SECRET_KEY = getenv("AWS_S3_SECRET_KEY")
EXPORT_EMAIL_TEMPLATE_ID = getenv("EXPORT_EMAIL_TEMPLATE_ID")
STATIC_EMAILS = literal_eval(getenv("STATIC_EMAILS", "['testing.communication@credgenics.com']"))

CELERY_VISIBILITY_SETTINGS_ENABLED = getenv("CELERY_VISIBILITY_SETTINGS_ENABLED", False)
CELERY_VISIBILITY_TIMEOUT = int(getenv("CELERY_VISIBILITY_TIMEOUT", "3600"))

NOTICE_COMPANIES_WITH_PROPERTY_ADDRESS_HANDLING = getenv("NOTICE_COMPANIES_WITH_PROPERTY_ADDRESS_HANDLING", "").split(
    ","
)
LITIGATION_REPORT_DIRECTORY = getenv("LITIGATION_REPORT_DIRECTORY")
API_USER_AUTHENTICATION_TOKEN = getenv("API_USER_AUTHENTICATION_TOKEN")



import time
import traceback
from quart import Quart, g, request, make_response
from . import settings
from .cg_celery import make_celery
import json
from cg_database import Postgres
import cg_logger as logger
from cg_redis import RedisCache
from . import routes
from .utils import BadEnvException, VerifyEnv
from elasticapm.contrib.quart import ElasticAPM
from quart_schema import QuartSchema, RequestSchemaValidationError
from pydantic.error_wrappers import ValidationError

app = Quart(__name__)
app.config.from_object(settings)
QuartSchema(
    app,
    redoc_ui_path=f"{settings.BASE_ROUTE}/redocs",
    swagger_ui_path=f"{settings.BASE_ROUTE}/docs",
    openapi_path=f"{settings.BASE_ROUTE}/openapi.json",
    version="1.0.0",
    title="Queue Service API Documentation",
)
celery = make_celery(app=app)


@app.before_serving
async def _init():
    logger.init()
    apm = ElasticAPM(app)
    app.logger.info("logger initialized")

    verify_envs = VerifyEnv().verify()
    if verify_envs[0] is True:
        app.logger.info("Queue config variables verified")
    else:
        app.logger.error("QUEUE CONFIGS NOT VERIFIED %s", str(verify_envs[1]))
        error = "unable to verify these queue configs " + str(verify_envs[1])
        raise BadEnvException(message=error)

    await _init_db()
    app.logger.info("db initialized")

    await _init_redis()
    app.logger.info("redis initialized")

    _register_blueprints()
    app.logger.info("routes registered")
    return


@app.before_request
def _before_request():
    g.request_id = logger.get_request_id()
    g.request_start_time = time.time_ns()
    return


@app.after_request
def _after_request(response):
    response = _add_global_headers(response)
    return logger.print_access_log(response)


@app.errorhandler(400)
async def _bad_request(exception):
    body = await request.body
    app.logger.error("%s, request.args: %s, request.body: %s", exception, dict(request.args), body)

    return {"message": "unable to process the request due to invalid syntax"}, 400


@app.errorhandler(404)
def _route_not_found(exception):
    app.logger.error("%s, route: '%s'", exception, request.url)

    return {"message": "requested route not found"}, 404


@app.errorhandler(408)
def _resource_request_timed_out(exception):
    app.logger.error(
        "resource request timed out, traceback: %s",
        traceback.extract_tb(exception.__traceback__),
    )

    return {"message": "resource request timed out"}, 408


@app.errorhandler(405)
def _resource_request_method_not_allowed(exception):
    app.logger.error(
        "resource request method not allowed, traceback: %s",
        traceback.extract_tb(exception.__traceback__),
    )

    return {"message": "Method not allowed"}, 405


@app.errorhandler(Exception)
def _unhandled_exception(exception):
    app.logger.error(
        "server._unhandled_exception: %s, traceback: %s",
        exception,
        traceback.extract_tb(exception.__traceback__),
    )
    return {"message": f"Error occurred {exception}"}, 500


@app.errorhandler(RequestSchemaValidationError)
async def handle_request_validation_error(error):
    if isinstance((error.validation_error), (ValidationError)):
        error = json.loads(error.validation_error.json())
        if error[0]["type"] == "value_error":
            error = error[0]["msg"]
            return await make_response(
                json.dumps(
                    {
                        "message": "failed",
                        "output": "invalid payload keys",
                        "data": {"errors": [{"error": error}]},
                    }
                ),
                400,
            )
        else:
            loc = error[0]["loc"]
            if "__root__" in loc:
                loc.remove("__root__")
            return await make_response(
                json.dumps(
                    {
                        "message": "failed",
                        "output": "invalid payload keys",
                        "data": {"errors": [{"error": error[0]["msg"], "key": error[0]["loc"]}]},
                    }
                ),
                400,
            )
    elif str(error.validation_error) == "type object argument after ** must be a mapping, not NoneType":
        return await make_response(
            json.dumps(
                {
                    "message": "failed",
                    "output": "Empty request body",
                }
            ),
            400,
        )
    else:
        error = str(error.validation_error)
        return await make_response(
            json.dumps(
                {
                    "message": "failed",
                    "output": str(error.replace("__init__() ", "")),
                }
            ),
            400,
        )


def _add_global_headers(response):
    response.headers["X-Frame-Options"] = "SAMEORIGIN"
    response.headers["X-XSS-Protection"] = "1;mode=block"
    response.headers["X-Content-Type-Options"] = "nosniff"
    response.headers["Strict-Transport-Security"] = "max-age=31536000"
    response.headers["X-Request-ID"] = str(g.request_id)
    return response


async def _init_db():
    db_conf = app.config.get("POSTGRES")
    app.db = Postgres(
        database=db_conf["NAME"],
        host=db_conf["HOST"],
        port=db_conf["PORT"],
        user=db_conf["USER"],
        password=db_conf["PASSWORD"],
        enable_read_replica=db_conf["ENABLE_DB_READ_REPLICA"],
        read_replica_host=db_conf["READ_REPLICA_DB_HOST"],
        read_replica_port=db_conf["READ_REPLICA_DB_PORT"],
    )
    await app.db.connect()
    app.logger.info(f"DB initialized")

    campaign_db_conf = app.config.get("CAMPAIGN_POSTGRES")
    app.campaign_db = Postgres(
        database=campaign_db_conf["NAME"],
        host=campaign_db_conf["HOST"],
        port=campaign_db_conf["PORT"],
        user=campaign_db_conf["USER"],
        password=campaign_db_conf["PASSWORD"],
        enable_read_replica=campaign_db_conf["ENABLE_DB_READ_REPLICA"],
        read_replica_host=campaign_db_conf["READ_REPLICA_DB_HOST"],
        read_replica_port=campaign_db_conf["READ_REPLICA_DB_PORT"],
    )
    await app.campaign_db.connect()
    app.logger.info(f"campaign db initialized")

    recovery_db_conf = app.config.get("RECOVERY_POSTGRES")
    app.recovery_db = Postgres(
        database=recovery_db_conf["NAME"],
        host=recovery_db_conf["HOST"],
        port=recovery_db_conf["PORT"],
        user=recovery_db_conf["USER"],
        password=recovery_db_conf["PASSWORD"],
        enable_read_replica=recovery_db_conf["ENABLE_DB_READ_REPLICA"],
        read_replica_host=recovery_db_conf["READ_REPLICA_DB_HOST"],
        read_replica_port=recovery_db_conf["READ_REPLICA_DB_PORT"],
    )
    await app.recovery_db.connect()
    app.logger.info(f"recovery db initialized")

    comm_db_conf = app.config.get("COMM_POSTGRES")
    app.logger.info(f"comm_db_conf:{comm_db_conf}")
    app.comm_db = Postgres(
        database=comm_db_conf["NAME"],
        host=comm_db_conf["HOST"],
        port=comm_db_conf["PORT"],
        user=comm_db_conf["USER"],
        password=comm_db_conf["PASSWORD"],
        enable_read_replica=comm_db_conf["ENABLE_DB_READ_REPLICA"],
        read_replica_host=comm_db_conf["READ_REPLICA_DB_HOST"],
        read_replica_port=comm_db_conf["READ_REPLICA_DB_PORT"],
    )
    await app.comm_db.connect()
    app.logger.info("communication db initialized")
    return


async def _init_redis():
    app.redis = RedisCache(f"""{app.config.get("APP_NAME")}_{app.config.get("ENV")}""")
    redis_config = app.config.get("REDIS")
    if not redis_config:
        raise Exception("REDIS config not found")
    await app.redis.connect(host=redis_config["HOST"], port=redis_config["PORT"])


def _register_blueprints():
    app.register_blueprint(routes.bp)



"""
report.py
Usage: Generate Report
"""
import uuid
import xlsxwriter
import io
import json
import requests
import boto3
import base64
import csv
import shortuuid
import logging
import pathlib
import datetime
import os
from botocore import config as boto_config
from botocore.exceptions import ClientError
from app.utils import get_s3_client, validate_email

# from app.tasks.notice.notice_callback.commons import upload_file_to_s3
from celery import shared_task
from http import HTTPStatus
from ..settings import (
    AI_COMPLETION_REPORT_SENDGRID_ID,
    COMMUNICATION_SERVICE_BASE_URL,
    QUEUE_SERVICE_BASE_URL,
    RECOVERY_SERVICE_BASE_URL,
    SENDGRID_API_KEY,
    SENDGRID_TEMPLATE_ID,
    CAMPAIGN_SENDGRID_TEMPLATE_ID,
    BATCH_SENDGRID_TEMPLATE_ID,
    AWS_ACCESS_KEY_ID,
    AWS_SECRET_ACCESS_KEY,
    AWS_DEFAULT_REGION,
    S3_BUCKET_NAME,
    REPORT_SENDGRID_TEMPLATE_ID,
    COMPLETION_REPORT_SENDGRID_ID,
    UI_SERVICE_BASE_URL,
    EXPORTS_BUCKET_NAME,
    BCC_EMAILS,
    COUNTRY_WISE_MAPPING,
)
from ..choices import (
    REQUEST_COMPLETION_SMS_TEMP_ID,
    TypeOfTaskChoices,
    QueueTypeChoices,
    RequestTypeChoices,
    PROD_ENV,
    AI_RULE_EMAIL,
    CampaignTriggerTypeChoices,
    CampaignDeliveryTypeChoices,
    THRESHOLD_PERCENTAGE,
    COMPLETION_PERCENTAGE,
    PRINCIPAL_ENTITY_ID,
    MODULE,
    SOURCE,
    ColumnMapChoices,
    CAMPAIGN_STOPPED_ERROR_CODE,
)
from ..utils import format_email_dict, get_campaign_summary, get_users_details
from ..service import (
    create_dynamic_template,
    create_batch_dynamic_template,
    trigger_sms_communication,
    get_campaign_report_data,
    trigger_email_communication,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
s3_v4_config = boto_config.Config(signature_version="s3v4")

QUEUE_FAILURE = "QUEUE_FAILURE"


@shared_task(name="calculate_report_data", bind=True, ignore_result=False)
def calculate_report_data(self, *args, **kwargs):
    logger.info(f"calculate_report_data{kwargs}")
    type_of_task = kwargs["type_of_task"]
    url = f"{QUEUE_SERVICE_BASE_URL}/get_batch_number?batch_number=all&batch_id={kwargs['batch_id']}"
    try:
        batch_data = requests.request(
            method="GET",
            url=url,
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(kwargs["user"]),
                "X-Request-ID": kwargs["request_id"],
            },
        )
        logger.info(f"get_batch_number.status_code :: {batch_data.status_code}")
    except Exception as e:
        logger.info(f"get_batch_number.exception :: {str(e)}")
        return False

    if batch_data.status_code not in (200, 201):
        logger.info(f"get_batch_number.text :: {batch_data.text}")
        return False

    batch_data = json.loads(batch_data.text)

    data_list = batch_data["data"]
    report_string = "\n"
    failed_loan_data = []
    failed_loan_ids_list = []
    success_loan_data = []
    failed_data = []

    if data_list != []:
        time_data = data_list[0]
        time_data = json.loads(time_data["extras"])
        triggered_time = time_data["triggered_time"]

    for data in data_list:
        batch_status = json.loads(data["batch_status"])
        failed = batch_status["failed"]
        success = batch_status["success"]
        failed_loan_data.extend(failed)
        success_loan_data.extend(success)

    if type_of_task == TypeOfTaskChoices.scrape_indiapost.value:
        kwargs["company"] = "Credgenics"
        for fail_data in failed_loan_data:
            reason = fail_data["reason"]
            loan_id = fail_data["loan_id"]
            status_code = fail_data["status_code"]
            company_id = fail_data["company_id"]
            tracking_id = fail_data["tracking_id"]

            failed_loan_ids_list.append(loan_id)
            failed_data.append((loan_id, company_id, tracking_id, reason, status_code))
            report_string = report_string + f"<p>{loan_id} --company_id::{company_id} -- reason :: {reason}<p>"
    else:
        organisation_details_url = f"{RECOVERY_SERVICE_BASE_URL}/organisation/details?company_id={kwargs['company_id']}"
        try:
            organisation_data = requests.request(
                method="GET",
                url=organisation_details_url,
                headers={
                    "Content-Type": "application/json",
                    "X-CG-User": json.dumps(kwargs["user"]),
                    "X-Request-ID": kwargs["request_id"],
                },
            )
        except Exception as e:
            logger.info(f"organisation_data.exception :: {str(e)}")
            return False

        if organisation_data.status_code not in (200, 201):
            logger.info(f"organisation_data.text :: {organisation_data.text}")
            return False
        organisation_data = json.loads(organisation_data.text)
        for fail_data in failed_loan_data:
            reason = fail_data["reason"]
            loan_id = fail_data["loan_id"]
            status_code = fail_data["status_code"]

            failed_loan_ids_list.append(loan_id)
            failed_data.append((loan_id, reason, status_code))
            report_string = report_string + f"<p>{loan_id} -- reason :: {reason}<p>"
        kwargs["company"] = organisation_data["output"]["trademark"]

    kwargs["report_string"] = report_string
    kwargs["failed_loan_ids_list"] = failed_loan_ids_list
    kwargs["success_loan_ids_list"] = success_loan_data
    kwargs["failed_data"] = failed_data
    kwargs["triggered_time"] = triggered_time
    send_email_report.apply_async(queue=kwargs["queue"], kwargs=kwargs)
    return True


@shared_task(bind=True, name="queue_report_attach")
def queue_report_attach(self, *args, **kwargs):
    headers = {
        "authorization": f"Bearer {SENDGRID_API_KEY}",
        "content-type": "application/json",
        "Accept": "application/json",
    }
    mail_data = {
        "from": kwargs["from_data"],
        "reply_to": {"email": "reports@credgenics.com"},
        "personalizations": [
            {
                "to": [{"email": kwargs["to_email"]}],
                "cc": [{"email": "demo.agent@credgenics.com", "name": "Demo Agent"}],
                "dynamic_template_data": kwargs["dynamic_template_data"],
            }
        ],
        "template_id": kwargs["sendgrid_template_id"],
    }
    logger.info(f"mail_data:{mail_data}")
    file_path = None
    if kwargs["failed_loan_ids_list"] > 0:
        logger.info(f"failed_loan_ids: {kwargs['failed_loan_ids_list']}")
        unique_mail_id = shortuuid.ShortUUID().random(length=8)
        logger.info(f"unique_id: {unique_mail_id}")
        failed_data = kwargs["failed_data"]
        type_of_task = kwargs["type_of_task"]
        if type_of_task == TypeOfTaskChoices.scrape_indiapost.value:
            fields = ["Loan Id", "Company Id", "Tracking Id", "Reason", "Status code"]
        else:
            fields = ["Loan Id", "Reason", "Status code"]
        filename = f"bulk_{type_of_task}_report_{datetime.datetime.now().strftime('%d-%m-%Y-%H-%M-%S')}.csv"

        with open(filename, "w") as csvfile:
            file_writer = csv.writer(csvfile)
            file_writer.writerow(fields)
            file_writer.writerows(failed_data)

        file_path = f"{pathlib.Path().absolute()}/{filename}"
        with open(file_path, "rb") as f:
            file_data = f.read()
            f.close()

        # Encode contents of file as Base 64
        encoded = base64.b64encode(file_data).decode()
        mail_data["attachments"] = [
            {
                "content": encoded,
                "content_id": unique_mail_id,
                "disposition": "inline",
                "filename": filename,
                "name": filename,
                "type": "csv",
            }
        ]
    response = requests.request(
        method="POST",
        url="https://api.sendgrid.com/v3/mail/send",
        data=json.dumps(mail_data),
        headers=headers,
    )
    logger.info(f"queue_report_attach :: {response.status_code}")
    if file_path:
        pathlib.Path(f"{file_path}").unlink()
    if response.status_code != HTTPStatus.ACCEPTED.value:
        logger.info(f"queue_report_attach :: {dir(response)}")
        logger.info(f"queue_report_attach :: {response.text}")
        return False
    return True


@shared_task(name="send_email_report", bind=True, ignore_result=False)
def send_email_report(self, *args, **kwargs):
    logger.info(f"send_email_report_kwargs:{kwargs}")
    author = kwargs["author"]
    company = kwargs["company"]
    batch_id = kwargs["batch_id"]
    total_number_of_loan_ids = kwargs["total_loans"]
    type_of_task = kwargs["type_of_task"].replace("_", " ")
    queue_env = self.app.conf.QUEUE_NAME.lower()

    success_loan_ids_list = kwargs["success_loan_ids_list"]
    failed_loan_ids_list = kwargs["failed_loan_ids_list"]
    failed_data = kwargs["failed_data"]
    triggered_time = kwargs["triggered_time"]
    linked_loan = kwargs.get("linked_loan", False)
    logger.info(f"queue name: {kwargs['queue']}")
    try:
        user_name = author.split("@")[0]
        if linked_loan == True:
            report_message = "Total linked loan Id's"
        else:
            report_message = "Total Loan accounts"
        sendgrid_template_id = f"{SENDGRID_TEMPLATE_ID}"
        dynamic_template_data = {
            "user_name": user_name,
            "type_of_task": type_of_task,
            "company": company,
            "total_number_of_loan_ids": total_number_of_loan_ids,
            "success_loan_ids_list": len(success_loan_ids_list),
            "failed_loan_ids_list": len(failed_loan_ids_list),
            "triggered_time": triggered_time,
            "batch_id": batch_id,
            "report_message": report_message,
            "env": queue_env if queue_env != PROD_ENV else None,
        }
        if kwargs.get("template_id", None):
            dynamic_template_data["template_id"] = kwargs.get("template_id")
        if kwargs.get("draft_id", None):
            dynamic_template_data["draft_id"] = kwargs.get("draft_id")
        if kwargs.get("template_name", None):
            dynamic_template_data["template_name"] = kwargs.get("template_name")
        if kwargs.get("rule_id", None) and kwargs.get("rule_name", None):
            dynamic_template_data["rule_id"] = kwargs.get("rule_id")
            dynamic_template_data["rule_name"] = kwargs.get("rule_name")
        if kwargs.get("applied_filter", None):
            filters = kwargs.get("applied_filter")
            filters = filters.replace("&", " ;&")
            filters = filters.split("&")
            dynamic_template_data["applied_filter"] = filters

        kwargs["failed_loan_ids_list"] = len(failed_loan_ids_list)
        kwargs["dynamic_template_data"] = dynamic_template_data
        kwargs["to_email"] = author
        kwargs["failed_data"] = failed_data
        kwargs["from_data"] = {
            "name": f"{type_of_task.capitalize()} Report",
            "email": "reports@credgenics.com",
        }
        kwargs["sendgrid_template_id"] = sendgrid_template_id
        queue_report_attach.apply_async(queue=kwargs["queue"], kwargs=kwargs)
    except Exception as e:
        logger.error(f"send_email_report.exception:{str(e)}")
        return False
    return True


################################################################################
######################## Queue Report ##############################################
################################################################################


@shared_task(bind=True, name="send_report")
def send_report(self, *args, **kwargs):
    logger.info("app.tasks.report.send_report")
    completion_report = kwargs.get("completion_report", False)
    queue_report = kwargs.get("queue_report", False)
    campaign_id = kwargs.get("campaign_id")
    cc_emails = [{"email": "demo.agent@credgenics.com", "name": "Demo Agent"}]
    trigger_stop_count = kwargs.get("trigger_stop_count")

    if kwargs.get("cc_emails"):
        cc_emails.extend(kwargs.get("cc_emails"))
    bcc_emails = kwargs.get("bcc_emails")
    dynamic_template_data = kwargs["dynamic_template_data"]
    if isinstance(kwargs["to_email"], list):
        email_to = []
        for email in kwargs["to_email"]:
            email_to.append({"email": email})
    else:
        email_to = [{"email": kwargs["to_email"]}]
    if dynamic_template_data.get("triggered_time"):
        dynamic_template_data["triggered_time"] = dynamic_template_data["triggered_time"].split(".")[0]
    mail_data = {
        "from": kwargs["from_data"],
        "reply_to": {"email": "reports@credgenics.com"},
        "personalizations": [
            {
                "to": email_to,
                "cc": cc_emails,
                "dynamic_template_data": dynamic_template_data,
            }
        ],
        "template_id": kwargs["sendgrid_template_id"],
    }
    if bcc_emails:
        mail_data["personalizations"][0]["bcc"] = bcc_emails
    filename = kwargs.get("filename", None)
    logger.debug(f"send_report.file_attachment: {filename}")
    if filename:
        unique_mail_id = shortuuid.ShortUUID().random(length=8)
        logger.info(f"send_report.unique_id: {unique_mail_id}")
        mail_data["attachments"] = [
            {
                "content": kwargs["encoded"],
                "content_id": unique_mail_id,
                "disposition": "inline",
                "filename": filename,
                "name": filename,
                "type": "csv",
            }
        ]
    logger.debug(f"send_report.mail_data:{mail_data}")
    mail_data["module"] = MODULE
    mail_data["source"] = SOURCE
    response_text, response_status_code = trigger_email_communication(mail_data)
    logger.debug(f"send_report.mail_data.response.status_code ::{response_status_code}")
    if filename:
        path = pathlib.Path(f"{pathlib.Path().absolute()}/{filename}")
        if path.exists():
            path.unlink()

    if response_status_code != HTTPStatus.OK.value:
        logger.error(f"send_report.mail_data.error ::{response_text}")
        return False
    if completion_report:
        total_loans = kwargs.get("total_loans")
        delivery_progress_count = kwargs.get("delivery_progress_count")
        delivery_status = kwargs.get("delivery_status")
        trigger_progress_count = f"{total_loans}/{total_loans}"
        trigger_status = CampaignTriggerTypeChoices.completed.value
        payload = {
            "table": "campaigns",
            "set_clause": {
                "email_report_sent": "email_report_sent",
                "delivery_status": "delivery_status",
                "delivery_progress_count": "delivery_progress_count",
                "trigger_status": "trigger_status",
                "trigger_progress_count": "trigger_progress_count",
                "trigger_stop_count": "trigger_stop_count",
            },
            "from_values": [
                f"('{campaign_id}', true, '{delivery_status}', '{delivery_progress_count}', '{trigger_status}', '{trigger_progress_count}', {trigger_stop_count})"
            ],
            "from_columns": [
                "campaign_id",
                "email_report_sent",
                "delivery_status",
                "delivery_progress_count",
                "trigger_status",
                "trigger_progress_count",
                "trigger_stop_count",
            ],
            "from_alias_name": "cd",
            "where": {"campaign_id": "campaign_id"},
        }
        update_data_kwargs = {
            "payload": payload,
            "user": kwargs["user"],
            "request_id": kwargs["request_id"],
        }
        update_data.apply_async(
            queue=f"{self.app.conf.QUEUE_NAME}_{QueueTypeChoices.campaign_update.value}",
            kwargs=update_data_kwargs,
        )
    elif queue_report and campaign_id:
        total_loans = dynamic_template_data.get("total_loans")
        trigger_progress_count = f"{total_loans}/{total_loans}"
        trigger_status = CampaignTriggerTypeChoices.completed.value
        payload = {
            "table": "campaigns",
            "set_clause": {
                "trigger_status": "trigger_status",
                "trigger_progress_count": "trigger_progress_count",
                "trigger_stop_count": "trigger_stop_count",
            },
            "from_values": [f"('{campaign_id}', '{trigger_status}', '{trigger_progress_count}', {trigger_stop_count})"],
            "from_columns": [
                "campaign_id",
                "trigger_status",
                "trigger_progress_count",
                "trigger_stop_count",
            ],
            "from_alias_name": "cd",
            "where": {"campaign_id": "campaign_id"},
        }
        update_data_kwargs = {
            "payload": payload,
            "user": kwargs["user"],
            "request_id": kwargs["request_id"],
        }
        update_data.apply_async(
            queue=f"{self.app.conf.QUEUE_NAME}_{QueueTypeChoices.campaign_update.value}",
            kwargs=update_data_kwargs,
        )

    return True


@shared_task(name="report_data", bind=True, ignore_result=False)
def report_data(self, *args, **kwargs):
    """
    Makes data for generating csv file and report template data
    """
    logger.info(f"app.tasks.report.report_data")
    payload = kwargs["payload"]
    channel = payload["channel"]
    failed_data = kwargs["failed_data"]
    dynamic_template_data = kwargs["dynamic_template_data"]
    campaign_id = dynamic_template_data.get("campaign_id", "")
    batch_id = dynamic_template_data.get("batch_id", "")
    logger.info(f"report_data.dynamic_template_data.campaign_id {campaign_id}")
    logger.info(f"report_data.dynamic_template_data.batch_id {batch_id}")
    try:
        if dynamic_template_data["failed"] > 0:
            if channel == TypeOfTaskChoices.indiapost_tracking.value:
                fields = [
                    "Loan Id",
                    "Company Id",
                    "Tracking Number",
                    "Reason",
                    "Status code",
                ]
            elif channel == TypeOfTaskChoices.indiapost_upload.value:
                fields = ["Loan Id", "Tracking Number", "Reason", "Status code"]
            else:
                fields = ["Loan Id", "Reason", "Status code"]
            if kwargs.get("type_of_task", ""):
                filename = (
                    f"{kwargs.get('type_of_task')}_report_{datetime.datetime.now().strftime('%d-%m-%Y-%H-%M-%S')}.csv"
                )
            else:
                filename = f"{channel}_report_{datetime.datetime.now().strftime('%d-%m-%Y-%H-%M-%S')}.csv"

            with open(filename, "w") as csvfile:
                file_writer = csv.writer(csvfile)
                file_writer.writerow(fields)
                file_writer.writerows(failed_data)
            file_path = f"{pathlib.Path().absolute()}/{filename}"
            with open(file_path, "rb") as f:
                file_data = f.read()
                f.close()

            if channel == TypeOfTaskChoices.digital.value:
                s3_failure_report_excel_path = f"digital_notice_reports/reports/{batch_id}.csv"

                s3 = get_s3_client()

                if os.path.exists(file_path):
                    upload_file_to_s3(s3, file_path, S3_BUCKET_NAME, s3_failure_report_excel_path)

                    try:
                        url = f"{QUEUE_SERVICE_BASE_URL}/update_batch_response"
                        data = {
                            "batch_id": batch_id,
                            "s3_error_report_excel_path": s3_failure_report_excel_path,
                        }
                        headers = {
                            "Content-Type": "application/json",
                            "X-CG-User": json.dumps(payload["user"]),
                            "X-Request-ID": payload["request_id"],
                            "X-CG-Company": json.dumps(payload["company"]),
                        }

                        update_response = requests.request(method="PATCH", url=url, json=data, headers=headers)

                        logger.debug(
                            "update_batch_response.status_code:: %s",
                            update_response.status_code,
                        )

                        if update_response.status_code != 200:
                            logger.debug(
                                "update_batch_response.response:: %s",
                                update_response.text,
                            )

                    except Exception as e:
                        logger.error("report.report_data Exception: %s", e)

            # Encode contents of file as Base 64
            encoded = base64.b64encode(file_data).decode()
            kwargs["encoded"] = encoded
            kwargs["filename"] = filename
        kwargs["campaign_id"] = campaign_id
        kwargs["user"] = payload.get("user")
        kwargs["request_id"] = payload.get("request_id")
        kwargs["queue_report"] = True

        send_report.apply_async(
            queue=f"{self.app.conf.QUEUE_NAME}_{QueueTypeChoices.result.value}",
            kwargs=kwargs,
        )
    except Exception as e:
        logger.error(f"report_data.exception.error :: {str(e)}")
        return False
    return True


@shared_task(name="campaign_comm_export", bind=True, ignore_result=False)
def campaign_comm_export(self, *args, **kwargs):
    """
    {
        "campaign_id": "3c2e8ea080ef4f32a81740950085ee34",
        "channel": "email"
    }
    """
    logger.info(f"app.tasks.report.campaign_comm_export")
    company_id = kwargs["company_id"]
    channel = kwargs["channel"]
    campaign_id = kwargs["campaign_id"]
    queue = kwargs["queue"]
    is_deleted = kwargs.get("is_deleted", False)

    url = f"{COMMUNICATION_SERVICE_BASE_URL}/export/{channel}?company_id={company_id}"
    logger.debug(f"campaign_comm_export.export.url: {url}")
    try:
        response_data = {
            "campaign_id": campaign_id,
            "response_type": "data",
            "source": "campaign_report",
            "is_deleted": is_deleted,
        }
        campaign_comm_export = requests.request(
            method="POST",
            url=url,
            data=json.dumps(response_data),
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(kwargs["user"]),
                "X-Request-ID": kwargs["request_id"],
                "X-CG-Company": json.dumps(kwargs["company"]),
            },
        )
        logger.debug(f"campaign_comm_export.status_code :: {campaign_comm_export.status_code}")
    except Exception as e:
        logger.error(f"campaign_comm_export..exception :: {str(e)}")
        return False

    if campaign_comm_export.status_code != HTTPStatus.OK.value:
        logger.error(f"campaign_comm_export.error :: {campaign_comm_export.text}")
        return False

    #  0. create columns keys list

    #  1. export_comm keys none dict map with error source keys
    #  2. add error source keys to export comm ie. new_export_comm ### TODO list_1
    #  3. populate the export campaign_export data with 1 ### TODO list_2
    #  4. add list2 to list1

    campaign_comm_export = json.loads(campaign_comm_export.text)
    campaign_comm_export_output = campaign_comm_export.get("output", [])
    logger.info(f"campaign_comm_export_output :: {len(campaign_comm_export)}")
    if campaign_comm_export_output == []:
        campaign_comm_export_output = [ColumnMapChoices[channel].value.copy()]

    kwargs["export_comm"] = campaign_comm_export_output
    logger.info(f"campaign_comm_export.campaign_comm_export_output:{campaign_comm_export_output}")
    campaign_export.apply_async(queue=queue, kwargs=kwargs)
    return True


@shared_task(name="campaign_export", bind=True, ignore_result=False)
def campaign_export(self, *args, **kwargs):
    logger.info(f"app.tasks.report.campaign_export")
    logger.debug(f"campaign_export.kwargs {kwargs}")
    export_comm = kwargs["export_comm"]
    channel = kwargs["channel"]
    campaign_id = kwargs["campaign_id"]
    queue = kwargs["queue"]
    is_deleted = kwargs.get("is_deleted", False)
    company_id = kwargs["company_id"]
    url = f"{QUEUE_SERVICE_BASE_URL}/campaign_data"
    try:
        response_data = {
            "campaign_id": campaign_id,
            "channel": channel,
            "is_deleted": is_deleted,
            "profile_id": kwargs.get("profile_id")
            if kwargs.get("profile_id")
            else kwargs["user"].get("permission_group_id"),
            "masking": True,
        }
        campaign_export = requests.request(
            method="POST",
            url=url,
            data=json.dumps(response_data),
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(kwargs["user"]),
                "X-CG-Company": json.dumps(kwargs["company"]),
                "X-Request-ID": kwargs["request_id"],
            },
        )
        logger.debug(f"campaign_export.status_code :: {campaign_export.status_code}")
    except Exception as e:
        logger.error(f"campaign_export.exception :: {str(e)}")
        return False

    if campaign_export.status_code != HTTPStatus.OK.value:
        logger.error(f"campaign_export.error :: {campaign_export.text}")
        return False
    #  0. columns keys list
    campaign_export = json.loads(campaign_export.text)
    campaign_export_output = campaign_export.get("output", None)
    campaign_output_data = campaign_export_output["campaign_data"]
    campaign_output_data = campaign_output_data[0]
    logger.debug(f"campaign_export.campaign_output_data: {campaign_output_data}")
    campaign_output_comm_data = json.loads(campaign_output_data["data"])
    if not isinstance(campaign_output_comm_data, dict):
        campaign_output_comm_data = json.loads(campaign_output_comm_data)
    comm_failed_count = campaign_export_output.get("comm_failed_count", 0)
    attempted_live = campaign_export_output.get("attempted_live", False)
    final_consolidated_comm_export_data = get_campaign_report_data(
        kwargs,
        communication_export_data=export_comm,
        campaign_data=campaign_export_output,
    )
    s3_kwargs = {}
    s3_kwargs["payload"] = kwargs
    s3_kwargs["final_consolidated_comm_export_data"] = final_consolidated_comm_export_data
    s3_kwargs["columns_key_mapping"] = list(final_consolidated_comm_export_data[0].keys())
    s3_kwargs["triggered_time"] = campaign_output_data.get("created")
    s3_kwargs["filters"] = campaign_output_comm_data.get("filters")
    s3_kwargs["range"] = campaign_output_comm_data.get("range")
    s3_kwargs["failed_count"] = comm_failed_count
    s3_kwargs["attempted_live"] = attempted_live
    campaign_s3_upload.apply_async(queue=queue, kwargs=s3_kwargs)
    return True


@shared_task(name="campaign_s3_upload", bind=True, ignore_result=False)
def campaign_s3_upload(self, *args, **kwargs):
    logger.info(f"app.tasks.report.campaign_s3_upload")
    warning = False
    payload = kwargs["payload"]
    author = payload["author"]
    channel = payload["channel"]
    queue = payload["queue"]
    campaign_id = payload["campaign_id"]
    status = payload["delivery_status"]
    campaign_name = payload["name"]
    template_name = payload["template_name"]
    column_names = kwargs["columns_key_mapping"]
    completion_report = payload.get("completion_report", False)
    company_id = payload.get("company_id")
    company_trademark = payload["company"].get("trademark")
    rule_name = payload.get("rule_name")
    data = kwargs["final_consolidated_comm_export_data"]
    queue_env = self.app.conf.QUEUE_NAME.lower()
    is_deleted = payload.get("is_deleted", False)
    dc_notification_channels = payload["company"].get("dc_notification_channel", [])
    try:
        output = io.BytesIO()
        workbook = xlsxwriter.Workbook(output)
        worksheet = workbook.add_worksheet()
        cell_format = workbook.add_format({"bold": True})

        worksheet.write_row(0, 0, column_names, cell_format)

        for row_idx, values in enumerate(data):
            worksheet.write_row(row_idx + 1, 0, list(values.values()))

        workbook.close()
        output.seek(0)

        bucket_credentials = {
            "AWS_ACCESS_KEY_ID": AWS_ACCESS_KEY_ID,
            "AWS_SECRET_ACCESS_KEY": AWS_SECRET_ACCESS_KEY,
        }

        s3_link_uuid = str(uuid.uuid4().hex)
        file_name = f"campaigns-report/{s3_link_uuid}.xlsx"
        presigned_url = s3_upload_and_generate_presigned_url(
            bucket_credentials=bucket_credentials,
            bucket_name=EXPORTS_BUCKET_NAME,
            key=file_name,
            data_object=output,
        )
        logger.debug(f"campaign_s3_upload.presigned_url.{campaign_id} :: {presigned_url}")
        dynamic_template_data = {
            "channel": channel.replace("_", " ").capitalize(),
            "campaign_id": campaign_id,
            "campaign_name": campaign_name,
            "status": status,
            "template_name": template_name,
            "report_url": presigned_url,
            "env": queue_env if queue_env != PROD_ENV else None,
            "company": company_trademark,
            "triggered_time": kwargs.get("triggered_time"),
            "filters": kwargs.get("filters"),
            "range": kwargs.get("range"),
        }
        report_kwargs = {}
        if completion_report:
            report_kwargs["from_data"] = {
                "name": f"Bulk {channel.replace('_', ' ').capitalize()} Notification",
                "email": "reports@credgenics.com",
            }
            report_kwargs["delivery_progress_count"] = payload.get("delivery_progress_count")
            report_kwargs["delivery_status"] = payload.get("delivery_status")
            report_kwargs["total_loans"] = payload.get("total_loans")
            to_emails = author.split(",") if isinstance(author, str) else author
            dc_notification_email_ids = payload["company"].get("dc_notification_email_ids", [])
            valid_email_ids = []
            if dc_notification_email_ids:
                valid_email_ids = list(
                    filter(
                        lambda email: email not in to_emails and validate_email(email),
                        dc_notification_email_ids,
                    )
                )
                if valid_email_ids:
                    cc_emails = format_email_dict(valid_email_ids)
                    report_kwargs["cc_emails"] = cc_emails
                    logger.debug(f"cc_emails:- {cc_emails}")
            if BCC_EMAILS and isinstance(BCC_EMAILS, list):
                filtered_bcc_emails = list(
                    filter(lambda email: email not in to_emails and email not in valid_email_ids, BCC_EMAILS)
                )
                report_kwargs["bcc_emails"] = format_email_dict(filtered_bcc_emails)
                logger.debug(f"bcc_emails:- {report_kwargs['bcc_emails']}")
            dynamic_template_data[
                "campaign_link"
            ] = f"{UI_SERVICE_BASE_URL}/app/bulk-management?company_id={company_id}&bulk_id={campaign_id}"
            if rule_name:
                dynamic_template_data["rule_name"] = rule_name
                report_kwargs["sendgrid_template_id"] = str(AI_COMPLETION_REPORT_SENDGRID_ID)
            else:
                report_kwargs["sendgrid_template_id"] = str(COMPLETION_REPORT_SENDGRID_ID)
            try:
                export_comm_url = f"{COMMUNICATION_SERVICE_BASE_URL}/campaigns/count?company_id={company_id}&campaign_id={campaign_id}&channel={channel}&campaign_summary=True&is_deleted={is_deleted}"
                logger.debug(f"campaign_s3_upload.export_campaign_summary.url: {export_comm_url}")
                campaign_export = requests.request(
                    method="GET",
                    url=export_comm_url,
                    headers={
                        "Content-Type": "application/json",
                        "X-CG-User": json.dumps(payload["user"]),
                        "X-Request-ID": payload["request_id"],
                        "X-CG-Company": json.dumps(payload["company"]),
                    },
                )
                logger.debug(f"campaign_s3_upload.export_campaign_summary.status_code: {campaign_export.status_code}")

                if campaign_export.status_code != HTTPStatus.OK.value:
                    logger.error(f"campaign_summary.export_campaign_summary.error {campaign_export.text}")
                    return False
            except Exception as e:
                logger.error(f"campaign_summary.export_campaign_summary.exception: {str(e)}")
                return False
            campaign_export = json.loads(campaign_export.text)
            campaign_export_output = campaign_export.get("output", [])
            if not campaign_export_output:
                campaign_export_output = {}
            else:
                campaign_export_output = campaign_export_output[0]
            campaign_summary = get_campaign_summary(
                campaign_export_output,
                channel,
                failed_count=kwargs["failed_count"],
                attempted_live=kwargs["attempted_live"],
                completion_summary=True,
            )
            if campaign_summary and isinstance(campaign_summary, dict):
                camapign_summary_keys = []
                camapign_summary_values = campaign_summary.values()
                max_value = max(camapign_summary_values)
                camapign_summary_final_values = []
                camapign_summary_final_values_map = {}
                for key, value in campaign_summary.items():
                    camapign_summary_keys.append(key.capitalize())
                    if max_value == 0:
                        value_percentage = 0
                    else:
                        value_percentage = round((value / max_value) * 100)
                    if str(key).lower() == "delivered" and value_percentage < 50:
                        warning = True
                    camapign_summary_final_values.append(f"{value_percentage}% ({value})")
                    camapign_summary_final_values_map[key] = f"{value_percentage}% ({value})"
                dynamic_template_data["warning"] = warning
                if warning:
                    dynamic_template_data[
                        "subject"
                    ] = f"[Warning] Bulk {channel.replace('_', ' ').capitalize()} Request Completed : Delivery rate is less than 50%"
                else:
                    dynamic_template_data[
                        "subject"
                    ] = f"Bulk {channel.replace('_', ' ').capitalize()} Request Completed"
                dynamic_template_data["camapign_summary_keys"] = camapign_summary_keys
                dynamic_template_data["camapign_summary_values"] = camapign_summary_final_values
            if (
                payload["number"]
                and dc_notification_channels
                and (RequestTypeChoices.sms.value in dc_notification_channels)
            ):
                sms_report_kwargs = {}
                sms_report_kwargs["company_id"] = company_id
                sms_report_kwargs["campaign_name"] = campaign_name
                sms_report_kwargs["campaign_id"] = campaign_id
                sms_report_kwargs["campaign_summary"] = camapign_summary_final_values_map
                sms_report_kwargs["numbers"] = payload["number"]
                sms_report_kwargs["user"] = payload["user"]
                sms_report_kwargs["request_id"] = payload["request_id"]
                sms_report_kwargs["company"] = payload["company"]
                sms_report_kwargs["channel"] = channel.capitalize()
                sms_report_kwargs["delivery_progress_count"] = payload.get("delivery_progress_count")
                sms_report_kwargs["delivery_status"] = payload.get("delivery_status")
                sms_report_kwargs["total_loans"] = payload.get("total_loans")
                sms_report_kwargs["trigger_stop_count"] = payload.get("trigger_stop_count")
                send_sms_report.apply_async(queue=queue, kwargs=sms_report_kwargs)
        else:
            report_kwargs["sendgrid_template_id"] = str(REPORT_SENDGRID_TEMPLATE_ID)
            report_kwargs["from_data"] = {
                "name": f"Bulk {channel.replace('_', ' ').capitalize()} Report",
                "email": "reports@credgenics.com",
            }

        report_kwargs["dynamic_template_data"] = dynamic_template_data
        report_kwargs["to_email"] = author
        report_kwargs["campaign_id"] = campaign_id
        report_kwargs["completion_report"] = completion_report
        report_kwargs["user"] = payload["user"]
        report_kwargs["request_id"] = payload["request_id"]
        report_kwargs["trigger_stop_count"] = payload.get("trigger_stop_count")

        logger.debug(f"campaign_s3_upload.report_kwargs.{campaign_id} :: {report_kwargs}")
        if not completion_report:
            send_report.apply_async(queue=queue, kwargs=report_kwargs)
        elif (
            completion_report
            and dc_notification_channels
            and RequestTypeChoices.email.value in dc_notification_channels
        ):
            send_report.apply_async(queue=queue, kwargs=report_kwargs)
    except Exception as e:
        logger.error(f"campaign_s3_upload.exception.{campaign_id} :: {str(e)}")
        return False
    return True


@shared_task(name="save_report_url", bind=True, ignore_result=False)
def save_report_url(self, *args, **kwargs):
    pass


def s3_upload_and_generate_presigned_url(bucket_credentials, bucket_name, key, data_object):
    logger.info(f"app.tasks.report.s3_upload_and_generate_presigned_url")
    logger.info(
        "s3_upload_and_generate_presigned_url.bucket_name: %s, s3_upload_and_generate_presigned_url.key: %s",
        bucket_name,
        key,
    )

    aws_access_key_id = bucket_credentials.get("AWS_ACCESS_KEY_ID", "")
    aws_secret_access_key = bucket_credentials.get("AWS_SECRET_ACCESS_KEY", "")

    try:
        s3_client_v4 = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name="ap-south-1",
            config=s3_v4_config,
            endpoint_url="https://s3."+AWS_DEFAULT_REGION+".amazonaws.com",
        )
        s3_client_v4.put_object(Bucket=bucket_name, Body=data_object, Key=key)

        presigned_url = s3_client_v4.generate_presigned_url(
            "get_object", Params={"Bucket": bucket_name, "Key": key}, ExpiresIn=86400
        )
    except ClientError as error:
        raise error

    return presigned_url


@shared_task(name="campaign_update", bind=True, ignore_result=False)
def campaign_update(self, *args, **kwargs):
    logger.info(f"app.tasks.report.campaign_update")
    company_id = kwargs["company_id"]
    campaign_data = kwargs["campaign_data"]
    campaign_export_output_map = kwargs["campaign_export_output_map"]
    update_list = []
    queue = kwargs["queue"]
    dc_notification_channels = kwargs["company"].get("dc_notification_channel", [])
    try:
        for data in campaign_data:
            campaign_id = data["campaign_id"]
            logger.debug(f"campaign_update.campaign_id: {campaign_id}")
            name = data["name"]
            channel = data["channel"]
            created = datetime.datetime.strptime(data["created"], "%Y-%m-%dT%H:%M:%S.%f")
            latest_triggered_time = data["latest_triggered_time"]
            author_id = data.get("author_id")
            campaign_author = data.get("author")
            old_trigger_status = data.get("trigger_status")
            old_trigger_progress_count = data.get("trigger_progress_count")
            old_delivery_status = data.get("delivery_status")
            old_delivery_progress_count = data.get("delivery_progress_count")
            old_trigger_stop_count = data.get("trigger_stop_count", 0)
            old_delivery_count = old_delivery_progress_count.split("/")[0]
            email_report_sent = data.get("email_report_sent")
            sms_report_sent = data.get("sms_report_sent")
            total_loans = int(data["total_loans"])
            table = f"campaign_{channel}"
            comm_data = json.loads(data["comm_data"])
            queue_processed = 0
            queue_fails = 0
            campaign_stopped_count = 0

            if channel != RequestTypeChoices.dtmf_ivr.value:
                if not isinstance(comm_data, dict):
                    comm_data = json.loads(comm_data)
                if comm_data.get("template_name"):
                    template_name = comm_data["template_name"]
                elif comm_data.get("comm_dict", {}).get("template_name"):
                    template_name = comm_data["comm_dict"]["template_name"]
                else:
                    template_name = ""
            else:
                template_name = comm_data.get("template_name", "")
            url = f"{QUEUE_SERVICE_BASE_URL}/campaign/count"
            payload = {"table": table, "campaign_id": campaign_id}
            logger.debug(f"campaign_update.get_campaign_count.url: {url}")
            campaign_res = requests.request(
                method="POST",
                url=url,
                data=json.dumps(payload),
                headers={
                    "Content-Type": "application/json",
                    "X-CG-User": json.dumps(kwargs["user"]),
                    "X-Request-ID": kwargs["request_id"],
                },
            )
            logger.debug(f"campaign_update.get_campaign_count.status_code :: {campaign_res.status_code}")
            logger.debug(f"campaign_update.get_campaign_count.response :: {campaign_res.text}")
            res_text = json.loads(campaign_res.text)
            if campaign_res.status_code == HTTPStatus.OK.value:
                campaign_count = res_text.get("output", [])
                if not (isinstance(campaign_count, list)):
                    continue
            else:
                continue

            logger.debug(f"campaign_update.campaigns_count_result: {campaign_count}")
            for campaigns_data_item in campaign_count:
                queue_processed += campaigns_data_item.get("count", 0)
                queue_fails += campaigns_data_item.get("count", 0) if campaigns_data_item.get("status") == "FAIL" else 0
                campaign_stopped_count += (
                    campaigns_data_item.get("count", 0)
                    if campaigns_data_item.get("error_code") == CAMPAIGN_STOPPED_ERROR_CODE
                    else 0
                )
            trigger_stop_count = queue_processed - campaign_stopped_count
            trigger_progress_count = f"{queue_processed}/{total_loans}"
            trigger_percentage = (queue_processed / total_loans) * 100
            if queue_processed != total_loans:
                if THRESHOLD_PERCENTAGE <= trigger_percentage < COMPLETION_PERCENTAGE:
                    trigger_status = CampaignTriggerTypeChoices.partial_completed.value
                elif trigger_percentage >= COMPLETION_PERCENTAGE:
                    trigger_status = CampaignTriggerTypeChoices.completed.value
                elif datetime.datetime.now() - created > datetime.timedelta(hours=24):
                    trigger_status = CampaignTriggerTypeChoices.hold.value
                else:
                    trigger_status = CampaignTriggerTypeChoices.in_process.value
            else:
                trigger_status = CampaignTriggerTypeChoices.completed.value
            campaign_export_output_data = campaign_export_output_map.get(campaign_id, None)
            webhook_success_comm = 0
            if campaign_export_output_data is not None:
                webhook_success_comm = campaign_export_output_data.get("webhook_success_count", 0)
            total_process_success_count = queue_fails + webhook_success_comm
            delivery_progress_count = f"{total_process_success_count}/{total_loans}"
            delivery_percentage = (total_process_success_count / total_loans) * 100
            if total_process_success_count != total_loans:
                if THRESHOLD_PERCENTAGE <= delivery_percentage < COMPLETION_PERCENTAGE:
                    delivery_status = CampaignDeliveryTypeChoices.partial_completed.value
                elif delivery_percentage >= COMPLETION_PERCENTAGE:
                    delivery_status = CampaignDeliveryTypeChoices.completed.value
                elif datetime.datetime.now() - created > datetime.timedelta(hours=48):
                    delivery_status = CampaignDeliveryTypeChoices.hold.value
                else:
                    delivery_status = CampaignDeliveryTypeChoices.updating.value
            else:
                delivery_status = CampaignDeliveryTypeChoices.completed.value
            if (
                old_trigger_status != CampaignTriggerTypeChoices.completed.value
                and trigger_progress_count != old_trigger_progress_count
            ):
                latest_triggered_time = datetime.datetime.now()
            if (
                delivery_status
                in (
                    CampaignDeliveryTypeChoices.completed.value,
                    CampaignDeliveryTypeChoices.partial_completed.value,
                )
                and not (email_report_sent or sms_report_sent)
                and dc_notification_channels
                and (
                    (RequestTypeChoices.sms.value in dc_notification_channels and not sms_report_sent)
                    or (RequestTypeChoices.email.value in dc_notification_channels and not email_report_sent)
                )
            ):
                author = None
                number = None
                rule_name = comm_data.get("rule_name", "")
                rule_id = comm_data.get("rule_id", "")
                if rule_name:
                    author_id = comm_data.get("ai_author_ids", "")
                logger.info(f"campaign_update.author_ids {author_id}")
                if author_id:
                    user_details = get_users_details(
                        company_id,
                        author_id,
                        kwargs["user"],
                        kwargs["request_id"],
                        kwargs["company"],
                    )
                    author = user_details.get("user_emails")
                    number = user_details.get("user_numbers")
                    profile_id = user_details.get("author_profile_id")
                if not author:
                    if rule_name:
                        author = AI_RULE_EMAIL
                    else:
                        author = campaign_author
                report_data = {
                    "campaign_id": campaign_id,
                    "channel": channel,
                    "author": author,
                    "number": number,
                    "company_id": company_id,
                    "name": name,
                    "delivery_status": delivery_status,
                    "delivery_progress_count": delivery_progress_count,
                    "total_loans": total_loans,
                    "template_name": template_name,
                    "user": kwargs["user"],
                    "request_id": kwargs["request_id"],
                    "completion_report": True,
                    "queue": queue,
                    "company": kwargs["company"],
                    "rule_name": rule_name,
                    "rule_id": rule_id,
                    "trigger_stop_count": trigger_stop_count,
                    "profile_id": profile_id,
                }
                campaign_comm_export.apply_async(queue=queue, kwargs=report_data)
            elif (
                trigger_status != old_trigger_status
                or trigger_progress_count != old_trigger_progress_count
                or delivery_status != old_delivery_status
                or delivery_progress_count != old_delivery_progress_count
                or trigger_stop_count != old_trigger_stop_count
            ) and total_process_success_count >= int(old_delivery_count):
                if latest_triggered_time:
                    update_value = f"""('{campaign_id}', '{trigger_status}', '{trigger_progress_count}', '{delivery_status}', '{delivery_progress_count}', {trigger_stop_count}, '{latest_triggered_time}'::timestamp)"""
                else:
                    update_value = f"""('{campaign_id}', '{trigger_status}', '{trigger_progress_count}', '{delivery_status}', '{delivery_progress_count}', {trigger_stop_count}, null::timestamp)"""
                update_list.append(update_value)
        if update_list:
            payload = {
                "table": "campaigns",
                "set_clause": {
                    "trigger_status": "trigger_status",
                    "trigger_progress_count": "trigger_progress_count",
                    "delivery_status": "delivery_status",
                    "delivery_progress_count": "delivery_progress_count",
                    "trigger_stop_count": "trigger_stop_count",
                    "latest_triggered_time": "latest_triggered_time",
                },
                "from_values": update_list,
                "from_columns": [
                    "campaign_id",
                    "trigger_status",
                    "trigger_progress_count",
                    "delivery_status",
                    "delivery_progress_count",
                    "trigger_stop_count",
                    "latest_triggered_time",
                ],
                "from_alias_name": "cd",
                "where": {"campaign_id": "campaign_id"},
            }
            update_data_kwargs = {
                "payload": payload,
                "user": kwargs["user"],
                "request_id": kwargs["request_id"],
            }
            update_data.apply_async(queue=queue, kwargs=update_data_kwargs)
    except Exception as e:
        logger.error(f"campaign_update.exception {str(e)}")
        return False
    return True


@shared_task(name="update_data", bind=True, ignore_result=False)
def update_data(self, *args, **kwargs):
    """
    Update the records of a table
    """
    logger.info(f"app.tasks.report.update_data")
    payload = kwargs["payload"]
    logger.debug(f"app.tasks.report.update_data.payload: {payload}")
    try:
        url = f"{QUEUE_SERVICE_BASE_URL}/update/data"
        logger.debug(f"update_data.url: {url}")
        update_res = requests.request(
            method="POST",
            url=url,
            data=json.dumps(payload),
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(kwargs["user"]),
                "X-Request-ID": kwargs["request_id"],
            },
        )
        logger.debug(f"update_data.status_code :: {update_res.status_code}")
        if update_res.status_code != HTTPStatus.OK.value:
            logger.debug(f"update_data.error :: {update_res.text}")
            return False
    except Exception as e:
        logger.error(f"update_data.exception :: {str(e)}")
        return False
    return True


@shared_task(name="campaign_result", bind=True, ignore_result=False)
def campaign_result(self, *args, **kwargs):
    """
    Aggregates the campaign data from database
    """
    logger.info(f"app.tasks.report.campaign_result")
    payload = kwargs["payload"]
    logger.info(f"campaign_result.campaign_id {payload['campaign_id']}")
    company_details = payload["company"]
    request_id = payload["request_id"]
    user = payload["user"]
    failed_data = []
    success_data = []
    updated_end_time_flag = kwargs["updated_end_time_flag"]
    queue_env = self.app.conf.QUEUE_NAME.lower()
    try:
        url = f"{QUEUE_SERVICE_BASE_URL}/campaign_data"
        data = requests.request(
            method="POST",
            url=url,
            data=json.dumps(
                {
                    "campaign_id": payload["campaign_id"],
                    "channel": payload["channel"],
                    "updated_end_time_flag": updated_end_time_flag,
                    "trigger_report_flag": True,
                }
            ),
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(user),
                "X-Request-ID": request_id,
                "X-CG-Company": json.dumps(company_details),
            },
        )
        logger.debug(f"campaign_result.campaign_data.status_code :: {data.status_code}")
    except Exception as e:
        logger.error(f"campaign_result.campaign_data.exception :: {str(e)}")
        return False

    if data.status_code != HTTPStatus.OK.value:
        logger.error(f"campaign_result.campaign_data.error :: {data.text}")
        return False

    try:
        data = json.loads(data.text)
        output = data["output"]
        if (not output["channel_campaign_data"]) or (not output["campaign_data"]):
            return False
        campaign_data = output["campaign_data"][0]
        channel_campaign_data = output["channel_campaign_data"]
        trigger_stop_count = 0
        for data in channel_campaign_data:
            response = data["response"]
            status_code = data["status_code"]
            error_code = data["error_code"]
            loan_id = data["loan_id"]
            if data["status"] == "FAIL":
                failed_data.append((loan_id, response, status_code))
            if data["status"] == "SUCCESS":
                success_data.append((loan_id, response, status_code))
            if error_code != CAMPAIGN_STOPPED_ERROR_CODE:
                trigger_stop_count += 1

        dynamic_template_data = create_dynamic_template(
            payload=payload,
            failed=len(failed_data),
            created=campaign_data["created"],
            success=len(success_data),
            queue_env=queue_env,
            campaign_name=campaign_data["name"],
        )
        logger.debug(f"dynamic_template_data: {dynamic_template_data}")
        if isinstance(dynamic_template_data, dict):
            kwargs["dynamic_template_data"] = dynamic_template_data
            kwargs["to_email"] = payload["author"]
            kwargs["failed_data"] = failed_data
            kwargs["trigger_stop_count"] = trigger_stop_count
            kwargs["from_data"] = {
                "name": f"Bulk {payload['channel'].replace('_', ' ').capitalize()} Notification",
                "email": "reports@credgenics.com",
            }
            kwargs["sendgrid_template_id"] = str(CAMPAIGN_SENDGRID_TEMPLATE_ID)
            kwargs["role"] = payload["role"]
            rule_name = dynamic_template_data.get("rule_name")
            author = None
            if rule_name:
                author_id = dynamic_template_data.pop("ai_author_ids", None)
                if author_id:
                    user_details = get_users_details(
                        company_details.get("company_id"), author_id, user, request_id, company_details
                    )
                    author = user_details.get("user_emails")

            if not author:
                if rule_name:
                    author = AI_RULE_EMAIL
                else:
                    author = kwargs["to_email"]
            to_emails = author.split(",") if isinstance(author, str) else author
            kwargs["to_email"] = to_emails
            dc_notification_email_ids = payload["company"].get("dc_notification_email_ids", [])
            valid_email_ids = []
            if dc_notification_email_ids:
                valid_email_ids = list(
                    filter(
                        lambda email: email not in to_emails and validate_email(email),
                        dc_notification_email_ids,
                    )
                )
                if valid_email_ids:
                    kwargs["cc_emails"] = format_email_dict(valid_email_ids)
                    logger.debug(f"cc_emails:- {kwargs['cc_emails']}")
            if BCC_EMAILS and isinstance(BCC_EMAILS, list):
                filtered_bcc_emails = list(
                    filter(lambda email: email not in to_emails and email not in valid_email_ids, BCC_EMAILS)
                )
                kwargs["bcc_emails"] = format_email_dict(filtered_bcc_emails)
                logger.debug(f"bcc_emails:- {kwargs['bcc_emails']}")
            report_data.apply_async(
                queue=f"{self.app.conf.QUEUE_NAME}_{QueueTypeChoices.result.value}",
                kwargs=kwargs,
            )
    except Exception as e:
        logger.error(f"campaign_result.exception :: {str(e)}")
        return False
    logger.info("campaign_result.close")
    return True


@shared_task(name="batch_result", bind=True, ignore_result=False)
def batch_result(self, *args, **kwargs):
    """
    Aggregates the campaign data from database
    """
    logger.info(f"app.tasks.report.batch_result")
    payload = kwargs["payload"]
    failed_data = []
    success_data = []
    logger.info(f"batch_result.batch_id: {payload['batch_id']}")
    queue_env = self.app.conf.QUEUE_NAME.lower()

    try:
        url = f"{QUEUE_SERVICE_BASE_URL}/batch_data"
        data = requests.request(
            method="POST",
            url=url,
            data=json.dumps(
                {
                    "batch_id": payload["batch_id"],
                    "channel": payload["channel"],
                    "type": payload["type"],
                }
            ),
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(payload["user"]),
                "X-Request-ID": payload["request_id"],
                "X-CG-Company": json.dumps(payload["company"]),
            },
        )
        logger.debug(f"batch_result.batch_data.status_code :: {data.status_code}")
    except Exception as e:
        logger.error(f"batch_result.batch_data.exception :: {str(e)}")
        return False

    if data.status_code != HTTPStatus.OK.value:
        logger.error(f"batch_result.batch_data.error :: {data.text}")
        return False

    try:
        data = json.loads(data.text)
        output = data["output"]
        if not output["batch_data"]:
            return False
        batch_data = output["batch_data"][0]
        type = batch_data.get("type", "")
        channel = batch_data.get("channel", "")
        batch_loans_data = output["batch_loans_data"]
        if type == TypeOfTaskChoices.scrape.value and channel == TypeOfTaskChoices.indiapost_upload.value:
            for data in batch_loans_data:
                response = data["response"]
                status_code = data["status_code"]
                loan_id = data["loan_id"]
                tracking_number = data.get("tracking_number", "")
                failed_data.append((loan_id, tracking_number, response, status_code))
        elif type == TypeOfTaskChoices.scrape.value and channel == TypeOfTaskChoices.indiapost_tracking.value:
            for data in batch_loans_data:
                response = data["response"]
                status_code = data["status_code"]
                loan_id = data["loan_id"]
                company_id = data.get("company_id", "")
                tracking_number = data.get("tracking_number", "")
                failed_data.append((loan_id, company_id, tracking_number, response, status_code))
        elif type == TypeOfTaskChoices.notice.value:
            for data in batch_loans_data:
                status = data["status"]
                response = data["response"]
                status_code = data["status_code"]
                loan_id = data["loan_id"] if type != TypeOfTaskChoices.notice.value else "'" + str(data["loan_id"])
                if status == "SUCCESS":
                    success_data.append((loan_id, response, status_code))
                else:
                    failed_data.append((loan_id, response, status_code))

            payload["success_data"] = success_data
        else:
            for data in batch_loans_data:
                response = data["response"]
                status_code = data["status_code"]
                loan_id = data["loan_id"] if type != TypeOfTaskChoices.notice.value else "'" + str(data["loan_id"])
                failed_data.append((loan_id, response, status_code))

        dynamic_template_data = create_batch_dynamic_template(
            payload=payload,
            failed=len(failed_data),
            created=batch_data["created"],
            queue_env=queue_env,
        )
        type_of_task = ""
        if type == TypeOfTaskChoices.notice.value:
            type_of_task = f"{channel}_{type}"
        elif type == TypeOfTaskChoices.scrape.value:
            type_of_task = f"{channel}"
        else:
            type_of_task = f"{type}"
        kwargs["type_of_task"] = type_of_task
        type_of_task = type_of_task.replace("_", " ")
        if isinstance(dynamic_template_data, dict):
            kwargs["dynamic_template_data"] = dynamic_template_data
            kwargs["to_email"] = payload["author"]
            kwargs["failed_data"] = failed_data
            kwargs["from_data"] = {
                "name": f"{type_of_task.capitalize()} Report",
                "email": "reports@credgenics.com",
            }
            kwargs["sendgrid_template_id"] = str(BATCH_SENDGRID_TEMPLATE_ID)

            report_data.apply_async(
                queue=f"{self.app.conf.QUEUE_NAME}_{QueueTypeChoices.result.value}",
                kwargs=kwargs,
            )
    except Exception as e:
        logger.error(f"batch_result.exception :: {str(e)}")
        return False
    return True


@shared_task(bind=True, name="send_sms_report")
def send_sms_report(self, *args, **kwargs):
    logger.info("app.tasks.report.send_sms_report")
    logger.debug(f"send_sms_report.kwargs: {kwargs}")
    company_id = kwargs.get("company_id")
    campaign_name = kwargs.get("campaign_name")
    channel = kwargs.get("channel")
    campaign_id = kwargs.get("campaign_id")
    campaign_summary = kwargs.get("campaign_summary")
    triggered = campaign_summary.get("triggered", 0)
    delivered = campaign_summary.get("delivered", 0)
    sms_mobile_data = kwargs.get("numbers")
    delivery_progress_count = kwargs.get("delivery_progress_count")
    delivery_status = kwargs.get("delivery_status")
    total_loans = kwargs.get("total_loans")
    company_details = kwargs["company"]
    trigger_progress_count = f"{total_loans}/{total_loans}"
    trigger_status = CampaignTriggerTypeChoices.completed.value
    trigger_stop_count = kwargs.get("trigger_stop_count")

    country_isd_code = None
    trademark = None
    try:
        trademark = company_details.get("trademark", "").capitalize()
        sms_body = (
            f"{trademark} bulk {channel.lower()} request {campaign_name} completed on {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}. Triggered: {triggered}, Delivered: {delivered}\n- Credgenics",
        )
        for country_code, value in sms_mobile_data.items():
            if not len(value):
                logger.error(f"send_sms_report.exception :: No mobile number found")
                continue
            country_mapping_data = COUNTRY_WISE_MAPPING.get(str(country_code.split("+")[-1]))
            if country_mapping_data:
                if country_mapping_data["dlt_check_applicable"]:
                    data = {
                        "sms_mobile": ",".join(value),
                        "sms_body": str(sms_body),
                        "company_id": company_id,
                        "content_template_id": REQUEST_COMPLETION_SMS_TEMP_ID,
                        "principal_entity_id": PRINCIPAL_ENTITY_ID,
                        "country_isd_code": str(country_code.split("+")[-1]),
                    }
                else:
                    data = {
                        "sms_mobile": ",".join(value),
                        "sms_body": str(sms_body),
                        "company_id": company_id,
                        "country_isd_code": str(country_code.split("+")[-1]),
                    }
            data["module"] = MODULE
            data["source"] = SOURCE
            response_text, response_status_code = trigger_sms_communication(data)
            logger.debug(f"trigger_sms_communication.response.status_code ::{response_status_code}")
            if response_status_code != HTTPStatus.OK.value:
                logger.error(f"trigger_sms_communication.response.status_code ::{response_text}")
                continue

        update_payload = {
            "table": "campaigns",
            "set_clause": {
                "sms_report_sent": "sms_report_sent",
                "delivery_status": "delivery_status",
                "delivery_progress_count": "delivery_progress_count",
                "trigger_status": "trigger_status",
                "trigger_progress_count": "trigger_progress_count",
                "trigger_stop_count": "trigger_stop_count",
            },
            "from_values": [
                f"('{campaign_id}', true, '{delivery_status}', '{delivery_progress_count}', '{trigger_status}', '{trigger_progress_count}', {trigger_stop_count})"
            ],
            "from_columns": [
                "campaign_id",
                "sms_report_sent",
                "delivery_status",
                "delivery_progress_count",
                "trigger_status",
                "trigger_progress_count",
                "trigger_stop_count",
            ],
            "from_alias_name": "cd",
            "where": {"campaign_id": "campaign_id"},
        }
        update_data_kwargs = {
            "payload": update_payload,
            "user": kwargs["user"],
            "request_id": kwargs["request_id"],
        }
        update_data.apply_async(
            queue=f"{self.app.conf.QUEUE_NAME}_{QueueTypeChoices.campaign_update.value}",
            kwargs=update_data_kwargs,
        )
        return True
    except Exception as e:
        logger.error(f"send_sms_report.exception :: {str(e)}")
        return False



"""
tasks.py
Usage: Contains all generic tasks
"""

import json
import requests
import logging
import uuid
from datetime import datetime

from celery import shared_task
from celery.result import AsyncResult
from celery import group, chord

# from app.tasks.notice.digital_notice import handle_generate_digital_notice
# from app.tasks.notice.physical_notice import generate_physical_notices

# from ..tasks.notice.notice_callback.physical_notice import (
#     generate_physical_notices_batch_callback,
# )

# TODO import this from celery_app
from ..settings import (
    DTMF_CHUNK_SIZE,
    QUEUE_SERVICE_BASE_URL,
)

from .execution import (
    communication,
    address_conversion,
    upload_data,
    upload_disposition,
    whatsapp_optin,
    whatsapp_optin,
    indiapost_upload,
    indiapost_tracking,
    lat_long_conversion,
    upload_c2c_disposition,
    communication_dtmf_ivr,
    add_ecourt_case,
    update_ecourt_case,
    fetch_ecourt_case_orders,
    update_approval_request,
    create_litigation_report_messages,
)
from .delegator import delegator

from .callback import batch_callback, callback
from ..utils import _make_batch, _dict_key_filter, list_chunker
from ..choices import TypeOfTaskChoices, CHANNELS, RequestTypeChoices


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@shared_task(name="testing", bind=True)
def testing(uuid):
    """
    Task testing
    """
    result = AsyncResult(uuid)
    logger.info(f"Validate email address: {uuid}")

    exc = result.get(propagate=False)
    print("Task {0} raised exception: {1!r}\n{2!r}".format(uuid, exc, result.traceback))


@shared_task(bind=True, name="task_test")
def task_test(_):
    """
    Task testing
    """
    logger.info(f"This is task testing from queue")
    return True


@shared_task(bind=True, name="new_task_test")
def new_task_test(_):
    """
    Task testing
    """
    print("This is new task testing from queue")
    return True


@shared_task(bind=True, name="error_handler")
def error_handler(uuid):
    """
    Error handle task
    """
    result = AsyncResult(uuid)
    exc = result.get(propagate=False)
    print("Task {0} raised exception: {1!r}\n{2!r}".format(uuid, exc, result.traceback))
    return True


@shared_task(bind=True, name="assign_optin_batch_queue")
def assign_optin_batch_queue(self, *args, **kwargs):
    batch_list = []
    batch_id = kwargs["batch_id"]
    user = kwargs.get("user")
    request_id = kwargs.get("request_id")
    queue = kwargs["queue"]
    data = kwargs["data"]
    loan_ids = data["loan_ids"]
    payload = _dict_key_filter(data, ["loan_ids"])
    company_id = data["company_id"]
    author = kwargs["author"]
    logger.info(f"payload :: {payload}")
    # type_of_task = queue.split('_')[1]
    if data.get("type_of_comm", "") == "dtmf_ivr":
        type_of_task = "dtmf_ivr"
    else:
        type_of_task = queue.split("_", 1)[1]

    total_loans = len(loan_ids)

    for batch in _make_batch(loan_ids, 2000):
        batch_list.append(batch)

    for index, loan_batch in enumerate(batch_list, 1):
        url = f"{QUEUE_SERVICE_BASE_URL}/initialize_monitor"
        try:
            res = requests.request(
                method="POST",
                url=url,
                data=json.dumps(
                    {
                        "batch_id": batch_id,
                        "batch_number": str(index),
                        "total_batches": len(batch_list),
                        "type_of_task": type_of_task,
                        "author": author,
                        "company_id": company_id,
                        "total_loans": total_loans,
                        "extras": {
                            "loan_ids": loan_batch,
                            "queue": queue,
                            "payload": payload,
                            "user": user,
                            "request_id": request_id,
                            "triggered_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        },
                    }
                ),
                headers={
                    "Content-Type": "application/json",
                    "X-CG-User": json.dumps(user),
                    "X-Request-ID": request_id,
                },
            )
            logger.info(f"initialize_monitor.status_code :: {res.status_code}")
        except Exception as e:
            logger.info(f"initialize_monitor.exception :: {str(e)}")
            return False

        if res.status_code not in (200, 201):
            logger.info(f"initialize_monitor.text :: {res.text}")
            return False
    delegator.apply_async(queue=queue, kwargs=kwargs)
    return True


@shared_task(bind=True, name="delegate")
def delegate(self, *args, **kwargs):
    """
    Delegates bulk operations to queue
    Delegates the signature tasks to queue using chord workflow
    """
    logger.info(f"app.tasks.tasks.delegate")
    logger.debug(f"delegate.kwargs: {kwargs}")
    tasks = []
    queue = kwargs["queue"]
    batch_id = kwargs["batch_id"]
    type = kwargs["type"]
    channel = kwargs["channel"]
    logger.info(f"delegate.batch_id {batch_id}")

    generic_data = json.loads(kwargs["loan_data"])
    next_approver_email = generic_data[0].get("next_approver_email") if isinstance(generic_data[0], dict) else ""
    payload = _dict_key_filter(kwargs, ["loan_data"])
    callback_payload = dict(payload)
    callback_payload["next_approver_email"] = next_approver_email

    for message in generic_data:
        task_kwargs = {
            "payload": payload,
            "message": message,
        }

        if type == TypeOfTaskChoices.scrape.value and channel == TypeOfTaskChoices.indiapost_tracking.value:
            tasks.append(indiapost_tracking.s(**task_kwargs))
        elif type == TypeOfTaskChoices.lat_long_conversion.value:
            tasks.append(lat_long_conversion.s(**task_kwargs))
        elif type == TypeOfTaskChoices.scrape.value and channel == TypeOfTaskChoices.indiapost_upload.value:
            tasks.append(indiapost_upload.s(**task_kwargs))
        elif type == TypeOfTaskChoices.upload_c2c_disposition.value:
            tasks.append(upload_c2c_disposition.s(**task_kwargs))
        elif type == TypeOfTaskChoices.optin.value:
            tasks.append(whatsapp_optin.s(**task_kwargs))
        elif type == TypeOfTaskChoices.ecourt_tracking.value and channel == TypeOfTaskChoices.add_ecourt_case.value:
            tasks.append(add_ecourt_case.s(**task_kwargs))
        elif type == TypeOfTaskChoices.ecourt_tracking.value and channel == TypeOfTaskChoices.update_ecourt_case.value:
            tasks.append(update_ecourt_case.s(**task_kwargs))
        elif (
            type == TypeOfTaskChoices.ecourt_tracking.value
            and channel == TypeOfTaskChoices.fetch_ecourt_case_orders.value
        ):
            tasks.append(fetch_ecourt_case_orders.s(**task_kwargs))
        elif type == TypeOfTaskChoices.litigation_approval_request.value:
            tasks.append(update_approval_request.s(**task_kwargs))

    data = json.loads(callback_payload["data"])
    data = _dict_key_filter(
        data,
        [
            "allocation_month",
            "creditline",
            "delivery_partner",
            "all_applicants",
            "co_applicant",
            "applicant_type",
            "notice_type",
            "source",
            "notify_type",
            "reminder_offset",
            "reminder_offset_type",
            "trigger_time",
        ],
    )
    callback_payload["data"] = json.dumps(data)

    logger.info("delegate.assign.chord: %s", batch_id)
    logger.debug(f"delegate.callback_data: {callback_payload}")

    chord(group(tasks), body=callback.s(queue=queue, payload=callback_payload)).apply_async(queue=queue)
    logger.info("delegate.assigned.chord: %s", batch_id)
    return True


@shared_task(bind=True, name="assign_generic_batch_queue")
def assign_generic_batch_queue(self, *args, **kwargs):
    """
    Assign generic batch to queue
    """
    batch_id = kwargs["batch_id"]
    queue = kwargs["queue"]
    type_of_task = kwargs["type_of_task"]
    if type_of_task == TypeOfTaskChoices.speedpost_upload.value:
        generic_data = kwargs["data"]["loan_data"]
        comm_data = _dict_key_filter(kwargs["data"], ["loan_data"])
        payload = _dict_key_filter(kwargs, ["data"])
        payload = {**payload, **comm_data}
    else:
        generic_data = kwargs["data"]
        payload = _dict_key_filter(kwargs, ["data"])
    total_loans = len(generic_data)
    url = f"{QUEUE_SERVICE_BASE_URL}/initialize_monitor"
    try:
        res = requests.request(
            method="POST",
            url=url,
            data=json.dumps(
                {
                    "batch_id": batch_id,
                    "type_of_task": type_of_task,
                    "batch_number": "1",
                    "total_batches": 1,
                    "author": payload["author"],
                    "company_id": payload["company_id"],
                    "total_loans": total_loans,
                    "extras": {
                        "loan_ids": generic_data,
                        "queue": queue,
                        "payload": payload,
                        "triggered_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    },
                }
            ),
            headers={
                "Content-Type": "application/json",
                "X-CG-User": json.dumps(payload["user"]),
                "X-Request-ID": payload["request_id"],
            },
        )
        logger.info(f"initialize_monitor.assign_generic_batch_queue.status_code :: {res.status_code}")
    except Exception as e:
        logger.info(f"initialize_monitor.assign_generic_batch_queue.exception :: {str(e)}")
        return False

    if res.status_code not in (200, 201):
        logger.info(f"initialize_monitor.assign_generic_batch_queue.text :: {res.text}")
        return False

    payload["batch_number"] = "1"
    payload["total_batches"] = 1
    payload["total_loans"] = total_loans
    payload["triggered_time"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    if type_of_task == TypeOfTaskChoices.address_conversion_lat_long.value:
        address_conversion_tasks = []
        for index, message in enumerate(generic_data):
            logger.info("assign_generic_batch_queue.index: %s", index)
            task_kwargs = {
                "payload": payload,
                "message": message,
            }
            address_conversion_tasks.append(address_conversion.s(**task_kwargs))

        chord(
            group(address_conversion_tasks),
            body=batch_callback.s(queue=queue, payload=payload),
        ).apply_async(queue=queue)
    elif type_of_task == TypeOfTaskChoices.speedpost_upload.value:
        upload_data_tasks = []
        for index, message in enumerate(generic_data):
            logger.info("assign_generic_batch_queue.index: %s", index)
            task_kwargs = {
                "payload": payload,
                "message": message,
            }
            upload_data_tasks.append(upload_data.s(**task_kwargs))

        chord(
            group(upload_data_tasks),
            body=batch_callback.s(queue=queue, payload=payload),
        ).apply_async(queue=queue)
    elif type_of_task == TypeOfTaskChoices.upload_c2c_disposition.value:
        upload_c2c_disposition = []
        for index, message in enumerate(generic_data):
            logger.info("assign_generic_batch_queue.index: %s", index)
            task_kwargs = {
                "payload": payload,
                "message": message,
            }
            upload_c2c_disposition.append(upload_disposition.s(**task_kwargs))

        chord(
            group(upload_c2c_disposition),
            body=batch_callback.s(queue=queue, payload=payload),
        ).apply_async(queue=queue)

    return True


@shared_task(bind=True, name="assign_report_download_queue")
def assign_report_download_queue(self, *args, **kwargs):
    """
    Assign report download to queue
    """
    logger.info(f"app.tasks.tasks.assign_report_download_queue")
    queue = kwargs["queue"]
    type_of_task = "_".join(queue.split("_")[1:])
    data = kwargs["data"]
    generic_data = data["loan_ids"]
    total_loans = len(generic_data)

    try:
        url = f"{QUEUE_SERVICE_BASE_URL}/initialize_monitor"
        request_body = {
            "batch_id": kwargs["batch_id"],
            "batch_number": "1",
            "total_batches": 1,
            "type_of_task": type_of_task,
            "author": kwargs["author"],
            "company_id": kwargs["company_id"],
            "total_loans": total_loans,
            "extras": {
                "token": kwargs["token"],
                "role": kwargs["role"],
                "queue": queue,
                "batch": data,
                "triggered_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            },
        }
        headers = {
            "Content-Type": "application/json",
            "X-Request-Id": str(kwargs["request_id"]),
            "X-CG-User": json.dumps(kwargs["user"]),
        }

        response = requests.request(
            method="POST",
            url=url,
            json=request_body,
            headers=headers,
        )

        logger.debug(f"initialize_monitor.status_code :: {response.status_code}")

    except Exception as e:
        logger.error(f"initialize_monitor.exception :: {str(e)}")
        return False

    if response.status_code not in (200, 201):
        logger.error(f"initialize_monitor.text :: {response.text}")
        return False

    task_kwargs = {
        "company_id": kwargs["company_id"],
        "author": kwargs["author"],
        "token": kwargs["token"],
        "role": kwargs["role"],
        "queue": queue,
        "batch_id": kwargs["batch_id"],
        "batch_number": "1",
        "total_batches": 1,
        "data": kwargs["data"],
        "report_name": kwargs["report_name"],
        "request_id": kwargs["request_id"],
        "user": kwargs["user"],
        "company": kwargs["company"],
    }

    create_litigation_report_messages.s(**task_kwargs).apply_async(queue=queue)
    return True


@shared_task(bind=True, name="delegate_campaign")
def delegate_campaign(self, *args, **kwargs):
    """
    Delegates campaigns to queue
    Delegates the signature tasks to queue using chord workflow
    Campaigns : Email, sms, voice, whatsapp, whatsapp_bot, dtmf_ivr
    """
    logger.info(f"app.tasks.tasks.delegate_campaign")
    logger.debug(f"delegate_campaign.kwargs: {kwargs}")
    tasks = []
    campaign_id = kwargs["campaign_id"]
    campaign_dict = {"campaign_id": campaign_id, "source": "queue"}
    generic_data = json.loads(kwargs["loan_data"])
    payload = _dict_key_filter(kwargs, ["loan_data", "additional_loan_data"])
    additional_loan_data = kwargs.get("additional_loan_data", {})
    task_kwargs = {"payload": payload}
    task_kwargs["queue"] = kwargs["queue"]
    channel = kwargs["channel"]
    try:
        if channel in CHANNELS:
            if channel == RequestTypeChoices.dtmf_ivr.value:
                for loan_ids in list_chunker(generic_data, DTMF_CHUNK_SIZE):
                    task_kwargs["loan_ids"] = loan_ids
                    additional_data = {
                        loan_id: {
                            **additional_loan_data.get(loan_id, {}),
                            **campaign_dict,
                        }
                        for loan_id in loan_ids
                    }
                    task_kwargs["additional_data"] = additional_data
                    tasks.append(communication_dtmf_ivr.s(**task_kwargs))
                logger.info(f"delegate_campaign.campaign_id.chord.assign.communication_dtmf_ivr: {campaign_id}")
            else:
                for loan_id in generic_data:
                    task_kwargs["loan_id"] = loan_id
                    task_kwargs["additional_loan_dict"] = additional_loan_data.get(loan_id, {})
                    tasks.append(communication.s(**task_kwargs))
                logger.info(f"delegate_campaign.campaign_id.chord.assign: {campaign_id}")
            logger.info(f"tasks.delegate_campaign.campaign_id.chord.assigned: {campaign_id}")
            chord_id = chord(group(tasks), body=callback.s(payload=payload)).apply_async(queue=kwargs["queue"])
    except Exception as e:
        logger.error(f"delegate_campaign.exception :: {str(e)}")
        return False
    return True
